# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from .core import Cardinality, AccessRights
from .function_app import FunctionApp, Function, DataType, AuthLevel
from .http import HttpMethod

__all__ = [
    'FunctionApp',
    'Function',
    'DataType',
    'AuthLevel',
    'Cardinality',
    'AccessRights',
    'HttpMethod'
]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import unittest

import azure.functions as func
import azure.functions.queue as azf_q


class TestQueue(unittest.TestCase):
    def test_QueueMessage_initialize_without_args(self):
        # given
        expected_id = None
        expected_body = b""
        expected_pop_receipt = None

        # when
        test_queue_message = func.QueueMessage()

        # then
        assert expected_id == test_queue_message.id
        assert expected_body == test_queue_message.get_body()
        assert expected_pop_receipt == test_queue_message.pop_receipt

    def test_QueueMessage_initialize_all_arguments(self):
        # given
        expected_id: str = "Identifier"
        expected_body: bytes = b"Body"
        expected_pop_receipt: str = "Pop Receipt"

        # when
        test_queue_message = func.QueueMessage(
            id=expected_id,
            body=expected_body,
            pop_receipt=expected_pop_receipt
        )

        # then
        assert expected_id == test_queue_message.id
        assert expected_body == test_queue_message.get_body()
        assert expected_pop_receipt == test_queue_message.pop_receipt

    def test_QueueMessage_get_json(self):
        # given
        test_body: str = '{"isJSON": "True"}'
        expected_body = json.loads(test_body)

        # when
        test_queue_message = func.QueueMessage(
            body=test_body
        )

        # then
        assert expected_body == test_queue_message.get_json()

    def test_QueueMessage_input_type(self):
        check_input_type = (
            azf_q.QueueMessageInConverter.check_input_type_annotation
        )
        self.assertTrue(check_input_type(func.QueueMessage))
        self.assertFalse(check_input_type(str))
        self.assertFalse(check_input_type(bytes))

    def test_QueueMessage_output_type(self):
        check_output_type = (
            azf_q.QueueMessageOutConverter.check_output_type_annotation
        )
        self.assertTrue(check_output_type(func.QueueMessage))
        self.assertTrue(check_output_type(str))
        self.assertTrue(check_output_type(bytes))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest
import json

from azure.functions.durable_functions import (
    OrchestrationTriggerConverter,
    EnitityTriggerConverter,
    ActivityTriggerConverter
)
from azure.functions._durable_functions import (
    OrchestrationContext,
    EntityContext
)
from azure.functions.meta import Datum

CONTEXT_CLASSES = [OrchestrationContext, EntityContext]
CONVERTERS = [OrchestrationTriggerConverter, EnitityTriggerConverter]


class TestDurableFunctions(unittest.TestCase):
    def test_context_string_body(self):
        body = '{ "name": "great function" }'
        for ctx in CONTEXT_CLASSES:
            context = ctx(body)
            self.assertIsNotNone(getattr(context, 'body', None))

            content = json.loads(context.body)
            self.assertEqual(content.get('name'), 'great function')

    def test_context_string_cast(self):
        body = '{ "name": "great function" }'
        for ctx in CONTEXT_CLASSES:
            context = ctx(body)
            self.assertEqual(str(context), body)

            content = json.loads(str(context))
            self.assertEqual(content.get('name'), 'great function')

    def test_context_bytes_body(self):
        body = '{ "name": "great function" }'.encode('utf-8')
        for ctx in CONTEXT_CLASSES:
            context = ctx(body)
            self.assertIsNotNone(getattr(context, 'body', None))

            content = json.loads(context.body)
            self.assertEqual(content.get('name'), 'great function')

    def test_context_bytes_cast(self):
        # TODO: this is just like the test above
        # (test_orchestration_context_bytes_body)
        body = '{ "name": "great function" }'.encode('utf-8')
        for ctx in CONTEXT_CLASSES:
            context = ctx(body)
            self.assertIsNotNone(getattr(context, 'body', None))

            content = json.loads(context.body)
            self.assertEqual(content.get('name'), 'great function')

    def test_trigger_converter(self):
        datum = Datum(value='{ "name": "great function" }',
                      type=str)
        for converter in CONVERTERS:
            otc = converter.decode(datum, trigger_metadata=None)
            content = json.loads(otc.body)
            self.assertEqual(content.get('name'), 'great function')

    def test_trigger_converter_type(self):
        datum = Datum(value='{ "name": "great function" }'.encode('utf-8'),
                      type=bytes)
        for converter in CONVERTERS:
            otc = converter.decode(datum, trigger_metadata=None)
            content = json.loads(otc.body)
            self.assertEqual(content.get('name'), 'great function')

    def test_trigger_check_good_annotation(self):

        for converter, ctx in zip(CONVERTERS, CONTEXT_CLASSES):
            self.assertTrue(
                converter.check_input_type_annotation(ctx)
            )

    def test_trigger_check_bad_annotation(self):
        for dt in (str, bytes, int):
            for converter in CONVERTERS:
                self.assertFalse(
                    converter.check_input_type_annotation(dt)
                )

    def test_trigger_has_implicit_return(self):
        for converter in CONVERTERS:
            self.assertTrue(
                converter.has_implicit_output()
            )

    def test_activity_trigger_inputs(self):
        # Activity Trigger only accept string type from durable extensions
        # It will be JSON deserialized into expected data type
        data = [
            {
                'input': Datum('sample', 'string'),
                'expected_value': 'sample',
                'expected_type': str
            },
            {
                'input': Datum('123', 'string'),
                'expected_value': 123,
                'expected_type': int
            },
            {
                'input': Datum('1234.56', 'string'),
                'expected_value': 1234.56,
                'expected_type': float
            },
            {
                'input': Datum('[ "do", "re", "mi" ]', 'string'),
                'expected_value': ["do", "re", "mi"],
                'expected_type': list
            },
            {
                'input': Datum('{ "number": "42" }', 'string'),
                'expected_value': {"number": "42"},
                'expected_type': dict
            }
        ]

        for datum in data:
            decoded = ActivityTriggerConverter.decode(
                data=datum['input'],
                trigger_metadata=None)
            self.assertEqual(decoded, datum['expected_value'])
            self.assertEqual(type(decoded), datum['expected_type'])

    def test_activity_trigger_encode(self):
        # Activity Trigger allow any JSON serializable as outputs
        # The return value will be carried back to the Orchestrator function
        data = [
            {
                'output': str('sample'),
                'expected_value': Datum('"sample"', 'json'),
            },
            {
                'output': int(123),
                'expected_value': Datum('123', 'json'),
            },
            {
                'output': float(1234.56),
                'expected_value': Datum('1234.56', 'json')
            },
            {
                'output': list(["do", "re", "mi"]),
                'expected_value': Datum('["do", "re", "mi"]', 'json')
            },
            {
                'output': dict({"number": "42"}),
                'expected_value': Datum('{"number": "42"}', 'json')
            }
        ]

        for datum in data:
            encoded = ActivityTriggerConverter.encode(
                obj=datum['output'],
                expected_type=type(datum['output']))
            self.assertEqual(encoded, datum['expected_value'])

    def test_activity_trigger_decode(self):
        # Activity Trigger allow inputs to be any JSON serializables
        # The input values to the trigger should be passed into arguments
        data = [
            {
                'input': Datum('sample_string', 'string'),
                'expected_value': str('sample_string')
            },
            {
                'input': Datum('"sample_json_string"', 'json'),
                'expected_value': str('sample_json_string')
            },
            {
                'input': Datum('{ "invalid": "json"', 'json'),
                'expected_value': str('{ "invalid": "json"')
            },
            {
                'input': Datum('true', 'json'),
                'expected_value': bool(True),
            },
            {
                'input': Datum('123', 'json'),
                'expected_value': int(123),
            },
            {
                'input': Datum('1234.56', 'json'),
                'expected_value': float(1234.56)
            },
            {
                'input': Datum('["do", "re", "mi"]', 'json'),
                'expected_value': list(["do", "re", "mi"])
            },
            {
                'input': Datum('{"number": "42"}', 'json'),
                'expected_value': dict({"number": "42"})
            }
        ]

        for datum in data:
            decoded = ActivityTriggerConverter.decode(
                data=datum['input'],
                trigger_metadata=None)
            self.assertEqual(decoded, datum['expected_value'])

    def test_activity_trigger_has_implicit_return(self):
        self.assertTrue(
            ActivityTriggerConverter.has_implicit_output()
        )



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import EVENT_GRID, EVENT_GRID_TRIGGER
from azure.functions.decorators.core import Trigger, DataType, OutputBinding


class EventGridTrigger(Trigger):

    @staticmethod
    def get_binding_name() -> str:
        return EVENT_GRID_TRIGGER

    def __init__(self,
                 name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        super().__init__(name=name, data_type=data_type)


class EventGridOutput(OutputBinding):

    @staticmethod
    def get_binding_name() -> str:
        return EVENT_GRID

    def __init__(self,
                 name: str,
                 topic_endpoint_uri: str,
                 topic_key_setting: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.topic_endpoint_uri = topic_endpoint_uri
        self.topic_key_setting = topic_key_setting
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import datetime
import json
import typing

from . import _abc


class QueueMessage(_abc.QueueMessage):
    """A Queue message object.

    :param str id:
        An optional string specifying the ID of the message.

    :param body:
        A string or bytes instance specifying the message body.

    :param str pop_receipt:
        An optional string containing the pop receipt token.
    """

    def __init__(self, *,
                 id: typing.Optional[str] = None,
                 body: typing.Optional[typing.Union[str, bytes]] = None,
                 pop_receipt: typing.Optional[str] = None) -> None:
        self.__id = id
        self.__body = b''
        self.__pop_receipt = pop_receipt

        if body is not None:
            self.__set_body(body)

    @property
    def id(self) -> typing.Optional[str]:
        """Message ID."""
        return self.__id

    @property
    def dequeue_count(self) -> typing.Optional[int]:
        """The number of times this message has been dequeued."""
        return None

    @property
    def expiration_time(self) -> typing.Optional[datetime.datetime]:
        """A datetime object with the message expiry time."""
        return None

    @property
    def insertion_time(self) -> typing.Optional[datetime.datetime]:
        """A datetime object with the message queue insertion time."""
        return None

    @property
    def time_next_visible(self) -> typing.Optional[datetime.datetime]:
        """A datetime object with the time the message will be visible next."""
        return None

    @property
    def pop_receipt(self) -> typing.Optional[str]:
        """The message pop receipt token as a string."""
        return self.__pop_receipt

    def __set_body(self, body):
        if isinstance(body, str):
            body = body.encode('utf-8')

        if not isinstance(body, (bytes, bytearray)):
            raise TypeError(
                f'response is expected to be either of '
                f'str, bytes, or bytearray, got {type(body).__name__}')

        self.__body = bytes(body)

    def get_body(self) -> bytes:
        """Return message content as bytes."""
        return self.__body

    def get_json(self) -> typing.Any:
        """Decode and return message content as a JSON object.

        :return:
            Decoded JSON data.

        :raises ValueError:
            when the body of the message does not contain valid JSON data.
        """
        return json.loads(self.__body)

    def __repr__(self) -> str:
        return (
            f'<azure.QueueMessage id={self.id} at 0x{id(self):0x}>'
        )



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.blob import BlobTrigger, BlobOutput, BlobInput
from azure.functions.decorators.core import BindingDirection, DataType


class TestBlob(unittest.TestCase):
    def test_blob_trigger_valid_creation(self):
        trigger = BlobTrigger(name="req",
                              path="dummy_path",
                              connection="dummy_connection",
                              data_type=DataType.UNDEFINED,
                              dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "blobTrigger")
        self.assertEqual(trigger.get_dict_repr(), {
            "type": "blobTrigger",
            "direction": BindingDirection.IN,
            'dummyField': 'dummy',
            "name": "req",
            "dataType": DataType.UNDEFINED,
            "path": "dummy_path",
            "connection": "dummy_connection"
        })

    def test_blob_input_valid_creation(self):
        blob_input = BlobInput(name="res",
                               path="dummy_path",
                               connection="dummy_connection",
                               data_type=DataType.UNDEFINED,
                               dummy_field="dummy")

        self.assertEqual(blob_input.get_binding_name(), "blob")
        self.assertEqual(blob_input.get_dict_repr(), {
            "type": "blob",
            "direction": BindingDirection.IN,
            'dummyField': 'dummy',
            "name": "res",
            "dataType": DataType.UNDEFINED,
            "path": "dummy_path",
            "connection": "dummy_connection"
        })

    def test_blob_output_valid_creation(self):
        blob_output = BlobOutput(name="res",
                                 path="dummy_path",
                                 connection="dummy_connection",
                                 data_type=DataType.UNDEFINED,
                                 dummy_field="dummy")

        self.assertEqual(blob_output.get_binding_name(), "blob")
        self.assertEqual(blob_output.get_dict_repr(), {
            "type": "blob",
            "direction": BindingDirection.OUT,
            'dummyField': 'dummy',
            "name": "res",
            "dataType": DataType.UNDEFINED,
            "path": "dummy_path",
            "connection": "dummy_connection"
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import azure.functions as func
import azure.functions.sql as sql
from azure.functions.meta import Datum
import json


class TestSql(unittest.TestCase):
    def test_sql_decode_none(self):
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=None, trigger_metadata=None)
        self.assertIsNone(result)

    def test_sql_decode_string(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "string")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'SqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'SqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'SqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'SqlRow item should have name test')

    def test_sql_decode_bytes(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """.encode(), "bytes")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'SqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'SqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'SqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'SqlRow item should have name test')

    def test_sql_decode_json(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'SqlRowList should be non-None')
        self.assertEqual(len(result),
                         1,
                         'SqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['id'],
                         '1',
                         'SqlRow item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test',
                         'SqlRow item should have name test')

    def test_sql_decode_json_name_is_null(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": null
        }
        """, "json")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result,
                             'SqlRowList itself should be non-None')
        self.assertEqual(len(result),
                         1,
                         'SqlRowList should have exactly 1 item')
        self.assertEqual(result[0]['name'],
                         None,
                         'Item in SqlRowList should be None')

    def test_sql_decode_json_multiple_entries(self):
        datum: Datum = Datum("""
        [
            {
                "id": "1",
                "name": "test1"
            },
            {
                "id": "2",
                "name": "test2"
            }
        ]
        """, "json")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result),
                         2,
                         'SqlRowList should have exactly 2 items')
        self.assertEqual(result[0]['id'],
                         '1',
                         'First SqlRowList item should have id 1')
        self.assertEqual(result[0]['name'],
                         'test1',
                         'First SqlRowList item should have name test1')
        self.assertEqual(result[1]['id'],
                         '2',
                         'First SqlRowList item should have id 2')
        self.assertEqual(result[1]['name'],
                         'test2',
                         'Second SqlRowList item should have name test2')

    def test_sql_decode_json_multiple_nulls(self):
        datum: Datum = Datum("[null]", "json")
        result: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result),
                         1,
                         'SqlRowList should have exactly 1 item')
        self.assertEqual(result[0],
                         None,
                         'SqlRow item should be None')

    def test_sql_encode_sqlrow(self):
        sqlRow = func.SqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """)
        datum = sql.SqlConverter.encode(obj=sqlRow, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         1,
                         'Encoded value should be list of length 1')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')

    def test_sql_encode_sqlrowlist(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        sqlRowList: func.SqlRowList = sql.SqlConverter.decode(
            data=datum, trigger_metadata=None)
        datum = sql.SqlConverter.encode(obj=sqlRowList, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         1,
                         'Encoded value should be list of length 1')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')

    def test_sql_encode_list_of_sqlrows(self):
        sqlRows = [
            func.SqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """),
            func.SqlRow.from_json("""
            {
                "id": "2",
                "name": "test2"
            }
            """)
        ]
        datum = sql.SqlConverter.encode(obj=sqlRows, expected_type=None)
        self.assertEqual(datum.type,
                         'json',
                         'Datum type should be JSON')
        self.assertEqual(len(datum.python_value),
                         2,
                         'Encoded value should be list of length 2')
        self.assertEqual(datum.python_value[0]['id'],
                         '1',
                         'id should be 1')
        self.assertEqual(datum.python_value[0]['name'],
                         'test',
                         'name should be test')
        self.assertEqual(datum.python_value[1]['id'],
                         '2',
                         'id should be 2')
        self.assertEqual(datum.python_value[1]['name'],
                         'test2',
                         'name should be test2')

    def test_sql_encode_list_of_str_raises(self):
        strList = [
            """
        {
            "id": "1",
            "name": "test"
        }
        """
        ]
        self.assertRaises(NotImplementedError,
                          sql.SqlConverter.encode,
                          obj=strList,
                          expected_type=None)

    def test_sql_encode_list_of_sqlrowlist_raises(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "test"
        }
        """, "json")
        sqlRowListList = [
            sql.SqlConverter.decode(
                data=datum, trigger_metadata=None)
        ]
        self.assertRaises(NotImplementedError,
                          sql.SqlConverter.encode,
                          obj=sqlRowListList,
                          expected_type=None)

    def test_sql_input_type(self):
        check_input_type = sql.SqlConverter.check_input_type_annotation
        self.assertTrue(check_input_type(func.SqlRowList),
                        'SqlRowList should be accepted')
        self.assertFalse(check_input_type(func.SqlRow),
                         'SqlRow should not be accepted')
        self.assertFalse(check_input_type(str),
                         'str should not be accepted')

    def test_sql_output_type(self):
        check_output_type = sql.SqlConverter.check_output_type_annotation
        self.assertTrue(check_output_type(func.SqlRowList),
                        'SqlRowList should be accepted')
        self.assertTrue(check_output_type(func.SqlRow),
                        'SqlRow should be accepted')
        self.assertFalse(check_output_type(str),
                         'str should not be accepted')

    def test_sqlrow_json(self):
        # Parse SqlRow from JSON
        sqlRow = func.SqlRow.from_json("""
        {
            "id": "1",
            "name": "test"
        }
        """)
        self.assertEqual(sqlRow['id'],
                         '1',
                         'Parsed SqlRow id should be 1')
        self.assertEqual(sqlRow['name'],
                         'test',
                         'Parsed SqlRow name should be test')

        # Parse JSON from SqlRow
        sqlRowJson = json.loads(func.SqlRow.to_json(sqlRow))
        self.assertEqual(sqlRowJson['id'],
                         '1',
                         'Parsed JSON id should be 1')
        self.assertEqual(sqlRowJson['name'],
                         'test',
                         'Parsed JSON name should be test')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
import json

from typing import Any, List

from . import meta

from ._kafka import AbstractKafkaEvent


class KafkaEvent(AbstractKafkaEvent):
    """A concrete implementation of Kafka event message type."""

    def __init__(self, *,
                 body: bytes,
                 trigger_metadata: typing.Mapping[str, meta.Datum] = None,
                 key: typing.Optional[str] = None,
                 offset: typing.Optional[int] = None,
                 partition: typing.Optional[int] = None,
                 topic: typing.Optional[str] = None,
                 timestamp: typing.Optional[str] = None,
                 headers: typing.Optional[list] = None) -> None:
        self.__body = body
        self.__trigger_metadata = trigger_metadata
        self.__key = key
        self.__offset = offset
        self.__partition = partition
        self.__topic = topic
        self.__timestamp = timestamp
        self.__headers = headers

        # Cache for trigger metadata after Python object conversion
        self._trigger_metadata_pyobj: typing.Optional[
            typing.Mapping[str, typing.Any]] = None

    def get_body(self) -> bytes:
        return self.__body

    @property
    def key(self) -> typing.Optional[str]:
        return self.__key

    @property
    def offset(self) -> typing.Optional[int]:
        return self.__offset

    @property
    def partition(self) -> typing.Optional[int]:
        return self.__partition

    @property
    def topic(self) -> typing.Optional[str]:
        return self.__topic

    @property
    def timestamp(self) -> typing.Optional[str]:
        return self.__timestamp

    @property
    def headers(self) -> typing.Optional[list]:
        return self.__headers

    @property
    def metadata(self) -> typing.Optional[typing.Mapping[str, typing.Any]]:
        if self.__trigger_metadata is None:
            return None

        if self._trigger_metadata_pyobj is None:
            self._trigger_metadata_pyobj = {}

            for k, v in self.__trigger_metadata.items():
                self._trigger_metadata_pyobj[k] = v.value

        return self._trigger_metadata_pyobj

    def __repr__(self) -> str:
        return (
            f'<azure.KafkaEvent '
            f'key={self.key} '
            f'partition={self.offset} '
            f'offset={self.offset} '
            f'topic={self.topic} '
            f'timestamp={self.timestamp} '
            f'at 0x{id(self):0x}>'
        )


class KafkaConverter(meta.InConverter, meta.OutConverter, binding='kafka'):
    @classmethod
    def check_input_type_annotation(cls, pytype) -> bool:
        valid_types = (KafkaEvent)

        return (
            meta.is_iterable_type_annotation(pytype, valid_types)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def check_output_type_annotation(cls, pytype) -> bool:
        valid_types = (str, bytes)
        return (
            meta.is_iterable_type_annotation(pytype, str)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def decode(
        cls, data: meta.Datum, *, trigger_metadata
    ) -> typing.Union[KafkaEvent, typing.List[KafkaEvent]]:
        data_type = data.type

        if data_type in ['string', 'bytes', 'json']:
            return cls.decode_single_event(data, trigger_metadata)

        elif data_type in ['collection_bytes', 'collection_string']:
            return cls.decode_multiple_events(data, trigger_metadata)

        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

    @classmethod
    def decode_single_event(cls, data: meta.Datum,
                            trigger_metadata) -> KafkaEvent:
        data_type = data.type

        if data_type in ['string', 'json']:
            body = data.value.encode('utf-8')

        elif data_type == 'bytes':
            body = data.value

        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

        return KafkaEvent(body=body)

    @classmethod
    def decode_multiple_events(cls, data: meta.Datum,
                               trigger_metadata) -> typing.List[KafkaEvent]:
        parsed_data: List[bytes] = []

        if data.type == 'collection_bytes':
            parsed_data = data.value.bytes

        elif data.type == 'collection_string':
            parsed_data = [
                d.encode('utf-8') for d in data.value.string
            ]

        return [KafkaEvent(body=pd) for pd in parsed_data]

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        raise NotImplementedError('Output bindings are not '
                                  'supported for Kafka')


class KafkaTriggerConverter(KafkaConverter,
                            binding='kafkaTrigger', trigger=True):

    @classmethod
    def decode(
        cls, data: meta.Datum, *, trigger_metadata
    ) -> typing.Union[KafkaEvent, typing.List[KafkaEvent]]:
        data_type = data.type

        if data_type in ['string', 'bytes', 'json']:
            return cls.decode_single_event(data, trigger_metadata)
        elif data_type in ['collection_bytes', 'collection_string']:
            return cls.decode_multiple_events(data, trigger_metadata)
        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

    @classmethod
    def decode_single_event(cls, data: meta.Datum,
                            trigger_metadata) -> KafkaEvent:
        data_type = data.type

        if data_type in ['string', 'json']:
            body = data.value.encode('utf-8')

        elif data_type == 'bytes':
            body = data.value

        else:
            raise NotImplementedError(
                f'unsupported event data payload type: {data_type}')

        return KafkaEvent(
            body=body,
            timestamp=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Timestamp', python_type=str),
            key=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Key', python_type=str),
            partition=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Partition', python_type=int),
            offset=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Offset', python_type=int),
            topic=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Topic', python_type=str),
            headers=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Headers', python_type=list),
            trigger_metadata=trigger_metadata
        )

    @classmethod
    def decode_multiple_events(cls, data: meta.Datum,
                               trigger_metadata) -> typing.List[KafkaEvent]:
        parsed_data: List[bytes] = []

        if data.type == 'collection_bytes':
            parsed_data = data.value.bytes

        elif data.type == 'collection_string':
            parsed_data = [
                d.encode('utf-8') for d in data.value.string
            ]

        timestamp_props = trigger_metadata.get('TimestampArray')
        key_props = trigger_metadata.get('KeyArray')
        partition_props = trigger_metadata.get('PartitionArray')
        offset_props = trigger_metadata.get('OffsetArray')
        topic_props = trigger_metadata.get('TopicArray')
        header_props = trigger_metadata.get('HeadersArray')

        parsed_timestamp_props: List[Any] = cls.get_parsed_props(
            timestamp_props, parsed_data)

        parsed_key_props = cls.get_parsed_props(
            key_props, parsed_data)

        parsed_partition_props = cls.get_parsed_props(
            partition_props, parsed_data)

        parsed_offset_props: List[Any] = []
        if offset_props is not None:
            parsed_offset_props = [v for v in offset_props.value.sint64]
            if len(parsed_offset_props) != len(parsed_data):
                raise AssertionError(
                    'Number of bodies and metadata mismatched')

        parsed_topic_props: List[Any]
        if topic_props is not None:
            parsed_topic_props = [v for v in topic_props.value.string]

        parsed_headers_props: List[Any]
        if header_props is not None:
            parsed_headers_list = cls.get_parsed_props(header_props,
                                                       parsed_data)
            parsed_headers_props = [v for v in parsed_headers_list]

        events = []

        for i in range(len(parsed_data)):
            event = KafkaEvent(
                body=parsed_data[i],
                timestamp=parsed_timestamp_props[i],
                key=cls._decode_typed_data(
                    parsed_key_props[i], python_type=str),
                partition=parsed_partition_props[i],
                offset=parsed_offset_props[i],
                topic=parsed_topic_props[i],
                headers=parsed_headers_props[i],
                trigger_metadata=trigger_metadata
            )
            events.append(event)

        return events

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        raise NotImplementedError('Output bindings are not '
                                  'supported for Kafka')

    @classmethod
    def get_parsed_props(
            cls, props: meta.Datum, parsed_data) -> List[Any]:
        parsed_props: List[Any] = []
        if props is not None:
            parsed_props = json.loads(props.value)
        if len(parsed_data) != len(parsed_props):
            raise AssertionError('Number of bodies and metadata mismatched')
        return parsed_props



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import COSMOS_DB_TRIGGER, COSMOS_DB
from azure.functions.decorators.core import BindingDirection, DataType
from azure.functions.decorators.cosmosdb import CosmosDBTrigger, \
    CosmosDBOutput, CosmosDBInput


class TestCosmosDB(unittest.TestCase):
    def test_cosmos_db_trigger_valid_creation(self):
        trigger = CosmosDBTrigger(name="req", database_name="dummy_db",
                                  collection_name="dummy_collection",
                                  connection_string_setting="dummy_str",
                                  leases_collection_throughput=1,
                                  checkpoint_interval=2,
                                  checkpoint_document_count=3,
                                  feed_poll_delay=4,
                                  lease_renew_interval=5,
                                  lease_acquire_interval=6,
                                  lease_expiration_interval=7,
                                  lease_collection_name='coll_name',
                                  lease_collection_prefix='prefix',
                                  lease_connection_string_setting='setting',
                                  lease_database_name='db',
                                  max_items_per_invocation=8,
                                  start_from_beginning=False,
                                  create_lease_collection_if_not_exists=False,
                                  preferred_locations="dummy_loc",
                                  data_type=DataType.UNDEFINED,
                                  dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "cosmosDBTrigger")
        self.assertEqual(trigger.get_dict_repr(),
                         {"checkpointDocumentCount": 3,
                          "checkpointInterval": 2,
                          "collectionName": "dummy_collection",
                          "connectionStringSetting": "dummy_str",
                          "createLeaseCollectionIfNotExists": False,
                          "dataType": DataType.UNDEFINED,
                          "databaseName": "dummy_db",
                          "direction": BindingDirection.IN,
                          'dummyField': 'dummy',
                          "feedPollDelay": 4,
                          "leaseAcquireInterval": 6,
                          "leaseCollectionName": 'coll_name',
                          "leaseCollectionPrefix": 'prefix',
                          "leaseConnectionStringSetting": 'setting',
                          "leaseDatabaseName": 'db',
                          "leaseExpirationInterval": 7,
                          "leaseRenewInterval": 5,
                          "leasesCollectionThroughput": 1,
                          "maxItemsPerInvocation": 8,
                          "name": "req",
                          "preferredLocations": "dummy_loc",
                          "startFromBeginning": False,
                          "type": COSMOS_DB_TRIGGER})

    def test_cosmos_db_output_valid_creation(self):
        output = CosmosDBOutput(name="req",
                                database_name="dummy_db",
                                collection_name="dummy_collection",
                                connection_string_setting="dummy_str",
                                create_if_not_exists=False,
                                collection_throughput=1,
                                use_multiple_write_locations=False,
                                data_type=DataType.UNDEFINED,
                                partition_key='key',
                                preferred_locations='locs',
                                dummy_field="dummy")

        self.assertEqual(output.get_binding_name(), "cosmosDB")
        self.assertEqual(output.get_dict_repr(),
                         {'collectionName': 'dummy_collection',
                          'collectionThroughput': 1,
                          'connectionStringSetting': 'dummy_str',
                          'createIfNotExists': False,
                          'dataType': DataType.UNDEFINED,
                          'databaseName': 'dummy_db',
                          'direction': BindingDirection.OUT,
                          'dummyField': 'dummy',
                          'name': 'req',
                          'partitionKey': 'key',
                          'preferredLocations': 'locs',
                          'type': COSMOS_DB,
                          'useMultipleWriteLocations': False})

    def test_cosmos_db_input_valid_creation(self):
        cosmosdb_input = CosmosDBInput(name="req", database_name="dummy_db",
                                       collection_name="dummy_collection",
                                       connection_string_setting="dummy_str",
                                       id="dummy_id",
                                       sql_query="dummy_query",
                                       partition_key="dummy_partitions",
                                       data_type=DataType.UNDEFINED,
                                       dummy_field="dummy")
        self.assertEqual(cosmosdb_input.get_binding_name(), "cosmosDB")
        self.assertEqual(cosmosdb_input.get_dict_repr(),
                         {'collectionName': 'dummy_collection',
                          'connectionStringSetting': 'dummy_str',
                          'dataType': DataType.UNDEFINED,
                          'databaseName': 'dummy_db',
                          'direction': BindingDirection.IN,
                          'dummyField': 'dummy',
                          'id': 'dummy_id',
                          'name': 'req',
                          'partitionKey': 'dummy_partitions',
                          'sqlQuery': 'dummy_query',
                          'type': COSMOS_DB})



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import EVENT_HUB_TRIGGER, EVENT_HUB
from azure.functions.decorators.core import BindingDirection, Cardinality, \
    DataType
from azure.functions.decorators.eventhub import EventHubTrigger, EventHubOutput


class TestEventHub(unittest.TestCase):
    def test_event_hub_trigger_valid_creation(self):
        trigger = EventHubTrigger(name="req",
                                  connection="dummy_connection",
                                  event_hub_name="dummy_event_hub",
                                  cardinality=Cardinality.ONE,
                                  consumer_group="dummy_group",
                                  data_type=DataType.UNDEFINED,
                                  dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "eventHubTrigger")
        self.assertEqual(trigger.get_dict_repr(),
                         {"cardinality": Cardinality.ONE,
                          "connection": "dummy_connection",
                          "consumerGroup": "dummy_group",
                          "dataType": DataType.UNDEFINED,
                          "direction": BindingDirection.IN,
                          'dummyField': 'dummy',
                          "eventHubName": "dummy_event_hub",
                          "name": "req",
                          "type": EVENT_HUB_TRIGGER})

    def test_event_hub_output_valid_creation(self):
        output = EventHubOutput(name="res",
                                event_hub_name="dummy_event_hub",
                                connection="dummy_connection",
                                data_type=DataType.UNDEFINED,
                                dummy_field="dummy")

        self.assertEqual(output.get_binding_name(), "eventHub")
        self.assertEqual(output.get_dict_repr(),
                         {'connection': 'dummy_connection',
                          'dataType': DataType.UNDEFINED,
                          'direction': BindingDirection.OUT,
                          'dummyField': 'dummy',
                          'eventHubName': 'dummy_event_hub',
                          'name': 'res',
                          'type': EVENT_HUB})



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import inspect
import re
from abc import ABCMeta
from enum import Enum
from json import JSONEncoder
from typing import TypeVar, Optional, Union, Iterable, Type, Callable

T = TypeVar("T", bound=Enum)
SNAKE_CASE_RE = re.compile(r'^([a-z]+\d*_[a-z\d_]*|_+[a-z\d]+[a-z\d_]*)$',
                           re.IGNORECASE)
WORD_RE = re.compile(r'^([a-z]+\d*)$', re.IGNORECASE)


class StringifyEnum(Enum):
    """This class output name of enum object when printed as string."""

    def __str__(self):
        return str(self.name)


class BuildDictMeta(type):
    def __new__(mcs, name, bases, dct):
        """BuildDictMeta will apply to every binding.
        It will apply :meth:`add_to_dict` decorator to :meth:`__init__` of
        every binding class to collect list of params to include in building
        json dictionary which corresponds to function.json in legacy app.
        It will also apply :meth:`skip_none` to :meth:`get_dict_repr` to
        enable json dictionary generated for every binding has non-empty
        value fields. It is needed for enabling binding param optionality.
        """
        cls = super().__new__(mcs, name, bases, dct)
        setattr(cls, '__init__',
                cls.add_to_dict(getattr(cls, '__init__')))
        setattr(cls, 'get_dict_repr',
                cls.skip_none(getattr(cls, 'get_dict_repr')))
        return cls

    @staticmethod
    def skip_none(func):
        def wrapper(*args, **kw):
            res = func(*args, **kw)
            return BuildDictMeta.clean_nones(res)

        return wrapper

    @staticmethod
    def add_to_dict(func: Callable):
        def wrapper(*args, **kwargs):
            if args is None or len(args) == 0:
                raise ValueError(
                    f'{func.__name__} has no args. Please ensure func is an '
                    f'object method.')

            func(*args, **kwargs)

            self = args[0]

            init_params = set(inspect.signature(func).parameters.keys())
            init_params.update(kwargs.keys())
            for key in kwargs.keys():
                if not hasattr(self, key):
                    setattr(self, key, kwargs[key])

            setattr(self, 'init_params', init_params)

        return wrapper

    @staticmethod
    def clean_nones(value):
        """
        Recursively remove all None values from dictionaries and lists,
        and returns
        the result as a new dictionary or list.
        """
        if isinstance(value, list):
            return [BuildDictMeta.clean_nones(x) for x in value if
                    x is not None]
        elif isinstance(value, dict):
            return {
                key: BuildDictMeta.clean_nones(val)
                for key, val in value.items()
                if val is not None
            }
        else:
            return value


class ABCBuildDictMeta(ABCMeta, BuildDictMeta):
    pass


def parse_singular_param_to_enum(param: Optional[Union[T, str]],
                                 class_name: Type[T]) -> Optional[T]:
    if param is None:
        return None
    if isinstance(param, str):
        try:
            return class_name[param.upper()]
        except KeyError:
            raise KeyError(
                f"Can not parse str '{param}' to {class_name.__name__}. "
                f"Allowed values are {[e.name for e in class_name]}")

    return param


def parse_iterable_param_to_enums(
        param_values: Optional[Union[Iterable[str], Iterable[T]]],
        class_name: Type[T]) -> Optional[Iterable[T]]:
    if param_values is None:
        return None

    try:
        return [class_name[value.upper()] if isinstance(value, str) else value
                for value in param_values]
    except KeyError:
        raise KeyError(
            f"Can not parse '{param_values}' to "
            f"Optional[Iterable[{class_name.__name__}]]. "
            f"Please ensure param all list elements exist in "
            f"{[e.name for e in class_name]}")


def to_camel_case(snake_case_str: str):
    if snake_case_str is None or len(snake_case_str) == 0:
        raise ValueError(
            f"Please ensure arg name {snake_case_str} is not empty!")

    if not is_snake_case(snake_case_str) and not is_word(snake_case_str):
        raise ValueError(
            f"Please ensure {snake_case_str} is a word or snake case "
            f"string with underscore as separator.")
    words = snake_case_str.split('_')
    return words[0] + ''.join(ele.title() for ele in words[1:])


def is_snake_case(input_string: str) -> bool:
    """
    Checks if a string is formatted as "snake case".
    A string is considered snake case when:
    - it's composed only by lowercase/uppercase letters and digits
    - it contains at least one underscore
    - it does not start with a number
    *Examples:*
    >>> is_snake_case('foo_bar_baz') # returns true
    >>> is_snake_case('foo') # returns false
    :param input_string: String to test.
    :return: True for a snake case string, false otherwise.
    """
    return SNAKE_CASE_RE.match(input_string) is not None


def is_word(input_string: str) -> bool:
    """
    Checks if a string is one word.
    A string is considered one word when:
    - it's composed only by lowercase/uppercase letters and digits
    - it does not start with a number
    *Examples:*
    >>> is_word('1foo') # returns false
    >>> is_word('foo_') # returns false
    >>> is_word('foo') # returns true
    :param input_string: String to test.
    :return: True for one word string, false otherwise.
    """
    return WORD_RE.match(input_string) is not None


class StringifyEnumJsonEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, StringifyEnum):
            return str(o)

        return super().default(o)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import azure.functions as func
import azure.functions.cosmosdb as cdb
from azure.functions.meta import Datum


class TestCosmosdb(unittest.TestCase):
    def test_cosmosdb_convert_none(self):
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=None, trigger_metadata=None)
        self.assertIsNone(result)

    def test_cosmosdb_convert_string(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "awesome_name"
        }
        """, "string")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], 'awesome_name')

    def test_cosmosdb_convert_bytes(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "awesome_name"
        }
        """.encode(), "bytes")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], 'awesome_name')

    def test_cosmosdb_convert_json(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": "awesome_name"
        }
        """, "json")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], 'awesome_name')

    def test_cosmosdb_convert_json_name_is_null(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": null
        }
        """, "json")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], None)

    def test_cosmosdb_convert_json_internal_fields_assigned(self):
        datum: Datum = Datum("""
        {
            "id": "1",
            "name": null,
            "_rid": "dummy12344",
            "_self": "7U4=/docs/gpU4AJcm7U4KAAAAAAAAAA==/",
            "_etag": "000-0500-0000-62598ff00000",
            "_attachments": "attachments/",
            "_ts": 1650036720
        }
        """, "json")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0]['name'], None)
        self.assertEqual(result[0]['_rid'], "dummy12344")
        self.assertEqual(result[0]['_self'],
                         "7U4=/docs/gpU4AJcm7U4KAAAAAAAAAA==/")
        self.assertEqual(result[0]['_etag'], "000-0500-0000-62598ff00000")
        self.assertEqual(result[0]['_attachments'], "attachments/")
        self.assertEqual(result[0]['_ts'], 1650036720)

    def test_cosmosdb_convert_json_multiple_entries(self):
        datum: Datum = Datum("""
        [
            {
                "id": "1",
                "name": "awesome_name"
            },
            {
                "id": "2",
                "name": "bossy_name"
            }
        ]
        """, "json")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 2)
        self.assertEqual(result[0]['name'], 'awesome_name')
        self.assertEqual(result[1]['name'], 'bossy_name')

    def test_cosmosdb_convert_json_multiple_nulls(self):
        datum: Datum = Datum("[null]", "json")
        result: func.DocumentList = cdb.CosmosDBConverter.decode(
            data=datum, trigger_metadata=None)
        self.assertIsNotNone(result)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], None)

    def test_cosmosdb_input_type(self):
        check_input_type = cdb.CosmosDBConverter.check_input_type_annotation
        self.assertTrue(check_input_type(func.DocumentList))
        self.assertFalse(check_input_type(func.Document))
        self.assertFalse(check_input_type(str))

    def test_cosmosdb_output_type(self):
        check_output_type = cdb.CosmosDBConverter.check_output_type_annotation
        self.assertTrue(check_output_type(func.DocumentList))
        self.assertTrue(check_output_type(func.Document))
        self.assertFalse(check_output_type(str))



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import TIMER_TRIGGER, HTTP_TRIGGER, \
    HTTP_OUTPUT, QUEUE, QUEUE_TRIGGER, SERVICE_BUS, SERVICE_BUS_TRIGGER, \
    EVENT_HUB, EVENT_HUB_TRIGGER, COSMOS_DB, COSMOS_DB_TRIGGER, BLOB, \
    BLOB_TRIGGER, EVENT_GRID_TRIGGER, EVENT_GRID
from azure.functions.decorators.core import DataType, AuthLevel, \
    BindingDirection, AccessRights, Cardinality
from azure.functions.decorators.function_app import FunctionApp
from azure.functions.decorators.http import HttpTrigger, HttpMethod
from azure.functions.decorators.timer import TimerTrigger
from tests.decorators.testutils import assert_json


class TestFunctionsApp(unittest.TestCase):
    def setUp(self):
        self.func_app = FunctionApp()

    def _get_user_function(self, app):
        funcs = app.get_functions()
        self.assertEqual(len(funcs), 1)
        return funcs[0]

    def test_route_is_function_name(self):
        app = self.func_app
        test_func_name = "dummy_function"

        @app.function_name(test_func_name)
        @app.route()
        def dummy_func():
            pass

        func = self._get_user_function(app)

        self.assertEqual(func.get_function_name(), test_func_name)
        self.assertTrue(isinstance(func.get_trigger(), HttpTrigger))
        self.assertTrue(func.get_trigger().route, test_func_name)

    def test_route_is_python_function_name(self):
        app = self.func_app

        @app.route()
        def dummy_func():
            pass

        func = self._get_user_function(app)

        self.assertEqual(func.get_function_name(), "dummy_func")
        self.assertTrue(isinstance(func.get_trigger(), HttpTrigger))
        self.assertTrue(func.get_trigger().route, "dummy_func")

    def test_route_is_custom(self):
        app = self.func_app

        @app.function_name("dummy_function")
        @app.route("dummy")
        def dummy_func():
            pass

        func = self._get_user_function(app)

        self.assertEqual(func.get_function_name(), "dummy_function")
        self.assertTrue(isinstance(func.get_trigger(), HttpTrigger))
        self.assertTrue(func.get_trigger().route, "dummy")

    def test_timer_trigger_default_args(self):
        app = self.func_app

        @app.schedule(arg_name="req", schedule="dummy_schedule")
        def dummy_func():
            pass

        func = self._get_user_function(app)
        self.assertEqual(func.get_function_name(), "dummy_func")
        assert_json(self, func, {
            "scriptFile": "function_app.py",
            "bindings": [
                {
                    "name": "req",
                    "type": TIMER_TRIGGER,
                    "direction": BindingDirection.IN,
                    "schedule": "dummy_schedule"
                }
            ]
        })

    def test_timer_trigger_full_args(self):
        app = self.func_app

        @app.schedule(arg_name="req", schedule="dummy_schedule",
                      run_on_startup=False, use_monitor=False,
                      data_type=DataType.STRING, dummy_field='dummy')
        def dummy():
            pass

        func = self._get_user_function(app)
        assert_json(self, func, {
            "scriptFile": "function_app.py",
            "bindings": [
                {
                    "name": "req",
                    "type": TIMER_TRIGGER,
                    "dataType": DataType.STRING,
                    "direction": BindingDirection.IN,
                    'dummyField': 'dummy',
                    "schedule": "dummy_schedule",
                    "runOnStartup": False,
                    "useMonitor": False
                }
            ]
        })

    def test_route_default_args(self):
        app = self.func_app

        @app.route()
        def dummy():
            pass

        func = self._get_user_function(app)
        assert_json(self, func, {
            "scriptFile": "function_app.py",
            "bindings": [
                {
                    "direction": BindingDirection.IN,
                    "type": HTTP_TRIGGER,
                    "name": "req",
                    "authLevel": AuthLevel.FUNCTION,
                    "route": "dummy"
                },
                {
                    "direction": BindingDirection.OUT,
                    "type": HTTP_OUTPUT,
                    "name": "$return"
                }
            ]
        })

    def test_route_with_all_args(self):
        app = self.func_app

        @app.route(trigger_arg_name='trigger_name', binding_arg_name='out',
                   methods=(HttpMethod.GET, HttpMethod.PATCH),
                   auth_level=AuthLevel.FUNCTION, route='dummy_route',
                   trigger_extra_fields={"dummy_field": "dummy"},
                   binding_extra_fields={"dummy_field": "dummy"})
        def dummy():
            pass

        func = self._get_user_function(app)
        assert_json(self, func, {
            "scriptFile": "function_app.py",
            "bindings": [
                {
                    "direction": BindingDirection.IN,
                    'dummyField': 'dummy',
                    "type": HTTP_TRIGGER,
                    "name": "trigger_name",
                    "authLevel": AuthLevel.FUNCTION,
                    "route": "dummy_route",
                    "methods": [
                        "GET", "PATCH"
                    ]
                },
                {
                    "direction": BindingDirection.OUT,
                    'dummyField': 'dummy',
                    "type": HTTP_OUTPUT,
                    "name": "out",
                }
            ]
        })

    def test_queue_default_args(self):
        app = self.func_app

        @app.queue_trigger(arg_name="req", queue_name="dummy_queue",
                           connection="dummy_conn")
        @app.write_queue(arg_name="out", queue_name="dummy_out_queue",
                         connection="dummy_out_conn")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": QUEUE,
                                         "name": "out",
                                         "queueName": "dummy_out_queue",
                                         "connection": "dummy_out_conn"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": QUEUE_TRIGGER,
                                         "name": "req",
                                         "queueName": "dummy_queue",
                                         "connection": "dummy_conn"
                                     }]})

    def test_queue_trigger(self):
        app = self.func_app

        @app.queue_trigger(arg_name="req", queue_name="dummy_queue",
                           connection="dummy_conn")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": QUEUE_TRIGGER,
            "name": "req",
            "queueName": "dummy_queue",
            "connection": "dummy_conn"
        })

    def test_queue_output_binding(self):
        app = self.func_app

        @app.queue_trigger(arg_name="req", queue_name="dummy_queue",
                           connection="dummy_conn")
        @app.write_queue(arg_name="out", queue_name="dummy_out_queue",
                         connection="dummy_out_conn")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": QUEUE,
            "name": "out",
            "queueName": "dummy_out_queue",
            "connection": "dummy_out_conn"
        })

    def test_queue_full_args(self):
        app = self.func_app

        @app.queue_trigger(arg_name="req", queue_name="dummy_queue",
                           connection="dummy_conn",
                           data_type=DataType.STRING, dummy_field="dummy")
        @app.write_queue(arg_name="out", queue_name="dummy_out_queue",
                         connection="dummy_out_conn",
                         data_type=DataType.STRING, dummy_field="dummy")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STRING,
                                         "type": QUEUE,
                                         "name": "out",
                                         "queueName": "dummy_out_queue",
                                         "connection": "dummy_out_conn"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STRING,
                                         "type": QUEUE_TRIGGER,
                                         "name": "req",
                                         "queueName": "dummy_queue",
                                         "connection": "dummy_conn"
                                     }]})

    def test_service_bus_queue_default_args(self):
        app = self.func_app

        @app.service_bus_queue_trigger(arg_name="req",
                                       connection="dummy_conn",
                                       queue_name="dummy_queue")
        @app.write_service_bus_queue(arg_name='res',
                                     connection='dummy_out_conn',
                                     queue_name='dummy_out_queue')
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": SERVICE_BUS,
                                         "name": "res",
                                         "connection": "dummy_out_conn",
                                         "queueName": "dummy_out_queue"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": SERVICE_BUS_TRIGGER,
                                         "name": "req",
                                         "connection": "dummy_conn",
                                         "queueName": "dummy_queue"
                                     }
                                 ]
                                 })

    def test_service_bus_queue_trigger(self):
        app = self.func_app

        @app.service_bus_queue_trigger(arg_name="req",
                                       connection="dummy_conn",
                                       queue_name="dummy_queue")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": SERVICE_BUS_TRIGGER,
            "name": "req",
            "connection": "dummy_conn",
            "queueName": "dummy_queue"
        })

    def test_service_bus_queue_output_binding(self):
        app = self.func_app

        @app.service_bus_queue_trigger(arg_name="req",
                                       connection="dummy_conn",
                                       queue_name="dummy_queue")
        @app.write_service_bus_queue(arg_name='res',
                                     connection='dummy_out_conn',
                                     queue_name='dummy_out_queue')
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": SERVICE_BUS,
            "name": "res",
            "connection": "dummy_out_conn",
            "queueName": "dummy_out_queue"
        })

    def test_service_bus_queue_full_args(self):
        app = self.func_app

        @app.service_bus_queue_trigger(arg_name="req",
                                       connection="dummy_conn",
                                       queue_name="dummy_queue",
                                       data_type=DataType.STREAM,
                                       access_rights=AccessRights.MANAGE,
                                       is_sessions_enabled=True,
                                       cardinality=Cardinality.MANY,
                                       dummy_field="dummy")
        @app.write_service_bus_queue(arg_name='res',
                                     connection='dummy_out_conn',
                                     queue_name='dummy_out_queue',
                                     data_type=DataType.STREAM,
                                     access_rights=AccessRights.MANAGE,
                                     dummy_field="dummy")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STREAM,
                                         "type": SERVICE_BUS,
                                         "name": "res",
                                         "connection": "dummy_out_conn",
                                         "queueName": "dummy_out_queue",
                                         "accessRights": AccessRights.MANAGE
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "dataType": DataType.STREAM,
                                         'dummyField': 'dummy',
                                         "type": SERVICE_BUS_TRIGGER,
                                         "name": "req",
                                         "connection": "dummy_conn",
                                         "queueName": "dummy_queue",
                                         "accessRights": AccessRights.MANAGE,
                                         "isSessionsEnabled": True,
                                         "cardinality": Cardinality.MANY
                                     }
                                 ]
                                 })

    def test_service_bus_topic_default_args(self):
        app = self.func_app

        @app.service_bus_topic_trigger(arg_name='req',
                                       connection='dummy_conn',
                                       topic_name='dummy_topic',
                                       subscription_name='dummy_sub')
        @app.write_service_bus_topic(arg_name='res', connection='dummy_conn',
                                     topic_name='dummy_topic',
                                     subscription_name='dummy_sub')
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "type": SERVICE_BUS,
                                         "direction": BindingDirection.OUT,
                                         "name": "res",
                                         "connection": "dummy_conn",
                                         "topicName": "dummy_topic",
                                         "subscriptionName": "dummy_sub"
                                     },
                                     {
                                         "type": SERVICE_BUS_TRIGGER,
                                         "direction": BindingDirection.IN,
                                         "name": "req",
                                         "connection": "dummy_conn",
                                         "topicName": "dummy_topic",
                                         "subscriptionName": "dummy_sub"
                                     }
                                 ]
                                 })

    def test_service_bus_topic_trigger(self):
        app = self.func_app

        @app.service_bus_topic_trigger(arg_name='req',
                                       connection='dummy_conn',
                                       topic_name='dummy_topic',
                                       subscription_name='dummy_sub')
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "type": SERVICE_BUS_TRIGGER,
            "direction": BindingDirection.IN,
            "name": "req",
            "connection": "dummy_conn",
            "topicName": "dummy_topic",
            "subscriptionName": "dummy_sub"
        })

    def test_service_bus_topic_output_binding(self):
        app = self.func_app

        @app.service_bus_topic_trigger(arg_name='req',
                                       connection='dummy_conn',
                                       topic_name='dummy_topic',
                                       subscription_name='dummy_sub')
        @app.write_service_bus_topic(arg_name='res', connection='dummy_conn',
                                     topic_name='dummy_topic',
                                     subscription_name='dummy_sub')
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "type": SERVICE_BUS,
            "direction": BindingDirection.OUT,
            "name": "res",
            "connection": "dummy_conn",
            "topicName": "dummy_topic",
            "subscriptionName": "dummy_sub"
        })

    def test_service_bus_topic_full_args(self):
        app = self.func_app

        @app.service_bus_topic_trigger(arg_name='req',
                                       connection='dummy_conn',
                                       topic_name='dummy_topic',
                                       subscription_name='dummy_sub',
                                       data_type=DataType.STRING,
                                       access_rights=AccessRights.LISTEN,
                                       is_sessions_enabled=False,
                                       cardinality=Cardinality.MANY,
                                       dummy_field="dummy")
        @app.write_service_bus_topic(arg_name='res', connection='dummy_conn',
                                     topic_name='dummy_topic',
                                     subscription_name='dummy_sub',
                                     data_type=DataType.STRING,
                                     access_rights=AccessRights.LISTEN,
                                     dummy_field="dummy")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "type": SERVICE_BUS,
                                         "direction": BindingDirection.OUT,
                                         'dummyField': 'dummy',
                                         "name": "res",
                                         "connection": "dummy_conn",
                                         "topicName": "dummy_topic",
                                         "subscriptionName": "dummy_sub",
                                         "dataType": DataType.STRING,
                                         "accessRights": AccessRights.LISTEN
                                     },
                                     {
                                         "type": SERVICE_BUS_TRIGGER,
                                         "direction": BindingDirection.IN,
                                         'dummyField': 'dummy',
                                         "name": "req",
                                         "connection": "dummy_conn",
                                         "topicName": "dummy_topic",
                                         "subscriptionName": "dummy_sub",
                                         "dataType": DataType.STRING,
                                         "accessRights": AccessRights.LISTEN,
                                         "isSessionsEnabled": False,
                                         "cardinality": Cardinality.MANY
                                     }
                                 ]
                                 })

    def test_event_hub_default_args(self):
        app = self.func_app

        @app.event_hub_message_trigger(arg_name="req",
                                       connection="dummy_connection",
                                       event_hub_name="dummy_event_hub")
        @app.write_event_hub_message(arg_name="res",
                                     event_hub_name="dummy_event_hub",
                                     connection="dummy_connection")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": EVENT_HUB,
                                         "name": "res",
                                         "connection": "dummy_connection",
                                         "eventHubName": "dummy_event_hub"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": EVENT_HUB_TRIGGER,
                                         "name": "req",
                                         "connection": "dummy_connection",
                                         "eventHubName": "dummy_event_hub"
                                     }
                                 ]
                                 })

    def test_event_hub_trigger(self):
        app = self.func_app

        @app.event_hub_message_trigger(arg_name="req",
                                       connection="dummy_connection",
                                       event_hub_name="dummy_event_hub")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": EVENT_HUB_TRIGGER,
            "name": "req",
            "connection": "dummy_connection",
            "eventHubName": "dummy_event_hub"
        })

    def test_event_hub_output_binding(self):
        app = self.func_app

        @app.event_hub_message_trigger(arg_name="req",
                                       connection="dummy_connection",
                                       event_hub_name="dummy_event_hub")
        @app.write_event_hub_message(arg_name="res",
                                     event_hub_name="dummy_event_hub",
                                     connection="dummy_connection")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": EVENT_HUB,
            "name": "res",
            "connection": "dummy_connection",
            "eventHubName": "dummy_event_hub"
        })

    def test_event_hub_full_args(self):
        app = self.func_app

        @app.event_hub_message_trigger(arg_name="req",
                                       connection="dummy_connection",
                                       event_hub_name="dummy_event_hub",
                                       cardinality=Cardinality.ONE,
                                       consumer_group="dummy_group",
                                       data_type=DataType.UNDEFINED,
                                       dummy_field="dummy")
        @app.write_event_hub_message(arg_name="res",
                                     event_hub_name="dummy_event_hub",
                                     connection="dummy_connection",
                                     data_type=DataType.UNDEFINED,
                                     dummy_field="dummy")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {
            "scriptFile": "function_app.py",
            "bindings": [
                {
                    "direction": BindingDirection.OUT,
                    'dummyField': 'dummy',
                    "dataType": DataType.UNDEFINED,
                    "type": EVENT_HUB,
                    "name": "res",
                    "connection": "dummy_connection",
                    "eventHubName": "dummy_event_hub"
                },
                {
                    "direction": BindingDirection.IN,
                    'dummyField': 'dummy',
                    "dataType": DataType.UNDEFINED,
                    "type": EVENT_HUB_TRIGGER,
                    "name": "req",
                    "connection": "dummy_connection",
                    "eventHubName": "dummy_event_hub",
                    "cardinality": Cardinality.ONE,
                    "consumerGroup": "dummy_group"
                }
            ]
        })

    def test_cosmosdb_full_args(self):
        app = self.func_app

        @app.cosmos_db_trigger(
            arg_name="trigger",
            database_name="dummy_db",
            collection_name="dummy_collection",
            connection_string_setting="dummy_str",
            lease_collection_name="dummy_lease_col",
            lease_connection_string_setting="dummy_lease_conn_str",
            lease_database_name="dummy_lease_db",
            leases_collection_throughput=1,
            lease_collection_prefix="dummy_lease_collection_prefix",
            checkpoint_interval=2,
            checkpoint_document_count=3,
            feed_poll_delay=4,
            lease_renew_interval=5,
            lease_acquire_interval=6,
            lease_expiration_interval=7,
            max_items_per_invocation=8,
            start_from_beginning=False,
            create_lease_collection_if_not_exists=False,
            preferred_locations="dummy_loc",
            data_type=DataType.STRING,
            dummy_field="dummy")
        @app.read_cosmos_db_documents(arg_name="in",
                                      database_name="dummy_in_db",
                                      collection_name="dummy_in_collection",
                                      connection_string_setting="dummy_str",
                                      id="dummy_id",
                                      sql_query="dummy_query",
                                      partition_key="dummy_partitions",
                                      data_type=DataType.STRING,
                                      dummy_field="dummy")
        @app.write_cosmos_db_documents(arg_name="out",
                                       database_name="dummy_out_db",
                                       collection_name="dummy_out_collection",
                                       connection_string_setting="dummy_str",
                                       create_if_not_exists=False,
                                       partition_key="dummy_part_key",
                                       collection_throughput=1,
                                       use_multiple_write_locations=False,
                                       preferred_locations="dummy_location",
                                       data_type=DataType.STRING,
                                       dummy_field="dummy")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STRING,
                                         "type": COSMOS_DB,
                                         "name": "out",
                                         "databaseName": "dummy_out_db",
                                         "collectionName":
                                             "dummy_out_collection",
                                         "connectionStringSetting":
                                             "dummy_str",
                                         "createIfNotExists": False,
                                         "collectionThroughput": 1,
                                         "useMultipleWriteLocations": False,
                                         "preferredLocations":
                                             "dummy_location",
                                         "partitionKey": "dummy_part_key"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STRING,
                                         "type": COSMOS_DB,
                                         "name": "in",
                                         "databaseName": "dummy_in_db",
                                         "collectionName":
                                             "dummy_in_collection",
                                         "connectionStringSetting":
                                             "dummy_str",
                                         "id": "dummy_id",
                                         "sqlQuery": "dummy_query",
                                         "partitionKey": "dummy_partitions"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         'dummyField': 'dummy',
                                         "dataType": DataType.STRING,
                                         "type": COSMOS_DB_TRIGGER,
                                         "name": "trigger",
                                         "databaseName": "dummy_db",
                                         "collectionName": "dummy_collection",
                                         "connectionStringSetting":
                                             "dummy_str",
                                         "leasesCollectionThroughput": 1,
                                         "checkpointInterval": 2,
                                         "checkpointDocumentCount": 3,
                                         "feedPollDelay": 4,
                                         "leaseRenewInterval": 5,
                                         "leaseAcquireInterval": 6,
                                         "leaseExpirationInterval": 7,
                                         "maxItemsPerInvocation": 8,
                                         "startFromBeginning": False,
                                         "createLeaseCollectionIfNotExists":
                                             False,
                                         "preferredLocations": "dummy_loc",
                                         "leaseCollectionName":
                                             "dummy_lease_col",
                                         "leaseConnectionStringSetting":
                                             "dummy_lease_conn_str",
                                         "leaseDatabaseName": "dummy_lease_db",
                                         "leaseCollectionPrefix":
                                             "dummy_lease_collection_prefix"
                                     }
                                 ]
                                 })

    def test_cosmosdb_default_args(self):
        app = self.func_app

        @app.cosmos_db_trigger(arg_name="trigger", database_name="dummy_db",
                               collection_name="dummy_collection",
                               connection_string_setting="dummy_str")
        @app.read_cosmos_db_documents(arg_name="in",
                                      database_name="dummy_in_db",
                                      collection_name="dummy_in_collection",
                                      connection_string_setting="dummy_str")
        @app.write_cosmos_db_documents(arg_name="out",
                                       database_name="dummy_out_db",
                                       collection_name="dummy_out_collection",
                                       connection_string_setting="dummy_str")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": COSMOS_DB,
                                         "name": "out",
                                         "databaseName": "dummy_out_db",
                                         "collectionName":
                                             "dummy_out_collection",
                                         "connectionStringSetting": "dummy_str"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": COSMOS_DB,
                                         "name": "in",
                                         "databaseName": "dummy_in_db",
                                         "collectionName":
                                             "dummy_in_collection",
                                         "connectionStringSetting": "dummy_str"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": COSMOS_DB_TRIGGER,
                                         "name": "trigger",
                                         "databaseName": "dummy_db",
                                         "collectionName": "dummy_collection",
                                         "connectionStringSetting": "dummy_str"
                                     }
                                 ]
                                 })

    def test_cosmosdb_trigger(self):
        app = self.func_app

        @app.cosmos_db_trigger(arg_name="trigger",
                               database_name="dummy_db",
                               collection_name="dummy_collection",
                               connection_string_setting="dummy_str")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": COSMOS_DB_TRIGGER,
            "name": "trigger",
            "databaseName": "dummy_db",
            "collectionName": "dummy_collection",
            "connectionStringSetting": "dummy_str"
        })

    def test_cosmosdb_input_binding(self):
        app = self.func_app

        @app.cosmos_db_trigger(arg_name="trigger",
                               database_name="dummy_db",
                               collection_name="dummy_collection",
                               connection_string_setting="dummy_str")
        @app.read_cosmos_db_documents(arg_name="in",
                                      database_name="dummy_in_db",
                                      collection_name="dummy_in_collection",
                                      connection_string_setting="dummy_str")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": COSMOS_DB,
            "name": "in",
            "databaseName": "dummy_in_db",
            "collectionName":
                "dummy_in_collection",
            "connectionStringSetting": "dummy_str"
        })

    def test_cosmosdb_output_binding(self):
        app = self.func_app

        @app.cosmos_db_trigger(arg_name="trigger",
                               database_name="dummy_db",
                               collection_name="dummy_collection",
                               connection_string_setting="dummy_str")
        @app.write_cosmos_db_documents(arg_name="out",
                                       database_name="dummy_out_db",
                                       collection_name="dummy_out_collection",
                                       connection_string_setting="dummy_str")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": COSMOS_DB,
            "name": "out",
            "databaseName": "dummy_out_db",
            "collectionName":
                "dummy_out_collection",
            "connectionStringSetting": "dummy_str"
        })

    def test_multiple_triggers(self):
        app = self.func_app
        with self.assertRaises(ValueError) as err:
            trigger1 = TimerTrigger(name="req1", schedule="dummy_schedule")
            trigger2 = TimerTrigger(name="req2", schedule="dummy_schedule")

            @app.schedule(arg_name="req1", schedule="dummy_schedule")
            @app.schedule(arg_name="req2", schedule="dummy_schedule")
            def dummy():
                pass
        self.assertEqual(err.exception.args[0],
                         "A trigger was already registered to this "
                         "function. Adding another trigger is not the "
                         "correct behavior as a function can only have one "
                         "trigger. Existing registered trigger "
                         f"is {trigger2.get_dict_repr()} and New trigger "
                         f"being added is {trigger1.get_dict_repr()}")

    def test_no_trigger(self):
        app = self.func_app
        with self.assertRaises(ValueError) as err:
            @app.write_queue(arg_name="out", queue_name="dummy_out_queue",
                             connection="dummy_out_conn")
            def dummy():
                pass

            app.get_functions()

        self.assertEqual(err.exception.args[0],
                         "Function dummy does not have a trigger. A valid "
                         "function must have one and only one trigger "
                         "registered.")

    def test_multiple_input_bindings(self):
        app = self.func_app

        @app.schedule(arg_name="req1", schedule="dummy_schedule")
        @app.read_cosmos_db_documents(
            arg_name="in1",
            database_name="dummy_in_db",
            collection_name="dummy_in_collection",
            connection_string_setting="dummy_str",
            id="dummy_id",
            sql_query="dummy_query",
            partition_key="dummy_partitions",
            data_type=DataType.STRING)
        @app.read_cosmos_db_documents(
            arg_name="in2",
            database_name="dummy_in_db",
            collection_name="dummy_in_collection",
            connection_string_setting="dummy_str",
            id="dummy_id",
            sql_query="dummy_query",
            partition_key="dummy_partitions",
            data_type=DataType.STRING)
        @app.write_queue(arg_name="out1", queue_name="dummy_out_queue",
                         connection="dummy_out_conn")
        @app.write_event_hub_message(
            arg_name="res",
            event_hub_name="dummy_event_hub",
            connection="dummy_connection")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": EVENT_HUB,
                                         "name": "res",
                                         "connection": "dummy_connection",
                                         "eventHubName": "dummy_event_hub"
                                     },
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": QUEUE,
                                         "name": "out1",
                                         "queueName": "dummy_out_queue",
                                         "connection": "dummy_out_conn"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "dataType": DataType.STRING,
                                         "type": COSMOS_DB,
                                         "name": "in2",
                                         "databaseName": "dummy_in_db",
                                         "collectionName":
                                             "dummy_in_collection",
                                         "connectionStringSetting":
                                             "dummy_str",
                                         "id": "dummy_id",
                                         "sqlQuery": "dummy_query",
                                         "partitionKey": "dummy_partitions"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "dataType": DataType.STRING,
                                         "type": COSMOS_DB,
                                         "name": "in1",
                                         "databaseName": "dummy_in_db",
                                         "collectionName":
                                             "dummy_in_collection",
                                         "connectionStringSetting":
                                             "dummy_str",
                                         "id": "dummy_id",
                                         "sqlQuery": "dummy_query",
                                         "partitionKey": "dummy_partitions"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": TIMER_TRIGGER,
                                         "name": "req1",
                                         "schedule": "dummy_schedule"
                                     }
                                 ]
                                 })

    def test_set_auth_level_for_http_functions(self):
        app = FunctionApp(http_auth_level=AuthLevel.ANONYMOUS)

        @app.route(auth_level=AuthLevel.ADMIN)
        def specify_auth_level():
            pass

        @app.route()
        def default_auth_level():
            pass

        funcs = app.get_functions()
        self.assertEqual(len(funcs), 2)
        http_func_1 = funcs[0]
        http_func_2 = funcs[1]

        self.assertEqual(http_func_1.get_user_function().__name__,
                         "specify_auth_level")
        self.assertEqual(http_func_2.get_user_function().__name__,
                         "default_auth_level")

        assert_json(self, http_func_1, {"scriptFile": "function_app.py",
                                        "bindings": [
                                            {
                                                "authLevel": AuthLevel.ADMIN,
                                                "type": HTTP_TRIGGER,
                                                "direction":
                                                    BindingDirection.IN,
                                                "name": "req",
                                                "route": "specify_auth_level"
                                            },
                                            {
                                                "type": HTTP_OUTPUT,
                                                "direction":
                                                    BindingDirection.OUT,
                                                "name": "$return"
                                            }
                                        ]
                                        })

        assert_json(self, http_func_2, {"scriptFile": "function_app.py",
                                        "bindings": [
                                            {
                                                "authLevel":
                                                    AuthLevel.ANONYMOUS,
                                                "type": HTTP_TRIGGER,
                                                "direction":
                                                    BindingDirection.IN,
                                                "name": "req",
                                                "route": "default_auth_level"
                                            },
                                            {
                                                "type": HTTP_OUTPUT,
                                                "direction":
                                                    BindingDirection.OUT,
                                                "name": "$return"
                                            }
                                        ]
                                        })

    def test_blob_default_args(self):
        app = self.func_app

        @app.blob_trigger(arg_name="req", path="dummy_path",
                          connection="dummy_conn")
        @app.read_blob(arg_name="file", path="dummy_path",
                       connection="dummy_conn")
        @app.write_blob(arg_name="out", path="dummy_out_path",
                        connection="dummy_out_conn")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func, {"scriptFile": "function_app.py",
                                 "bindings": [
                                     {
                                         "direction": BindingDirection.OUT,
                                         "type": BLOB,
                                         "name": "out",
                                         "path": "dummy_out_path",
                                         "connection": "dummy_out_conn"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": BLOB,
                                         "name": "file",
                                         "path": "dummy_path",
                                         "connection": "dummy_conn"
                                     },
                                     {
                                         "direction": BindingDirection.IN,
                                         "type": BLOB_TRIGGER,
                                         "name": "req",
                                         "path": "dummy_path",
                                         "connection": "dummy_conn"
                                     }]})

    def test_blob_trigger(self):
        app = self.func_app

        @app.blob_trigger(arg_name="req", path="dummy_path",
                          data_type=DataType.STRING,
                          connection="dummy_conn")
        def dummy():
            pass

        func = self._get_user_function(app)

        bindings = func.get_bindings()
        self.assertEqual(len(bindings), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.STRING,
            "type": BLOB_TRIGGER,
            "name": "req",
            "path": "dummy_path",
            "connection": "dummy_conn"
        })

    def test_blob_input_binding(self):
        app = self.func_app

        @app.blob_trigger(arg_name="req", path="dummy_path",
                          data_type=DataType.STRING,
                          connection="dummy_conn")
        @app.read_blob(arg_name="file", path="dummy_in_path",
                       connection="dummy_in_conn",
                       data_type=DataType.STRING)
        def dummy():
            pass

        func = self._get_user_function(app)

        bindings = func.get_bindings()
        self.assertEqual(len(bindings), 2)

        input_binding = bindings[0]
        output = bindings[1]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.STRING,
            "type": BLOB_TRIGGER,
            "name": "req",
            "path": "dummy_path",
            "connection": "dummy_conn"
        })

        self.assertEqual(input_binding.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.STRING,
            "type": BLOB,
            "name": "file",
            "path": "dummy_in_path",
            "connection": "dummy_in_conn"
        })

    def test_blob_output_binding(self):
        app = self.func_app

        @app.blob_trigger(arg_name="req", path="dummy_path",
                          data_type=DataType.STRING,
                          connection="dummy_conn")
        @app.write_blob(arg_name="out", path="dummy_out_path",
                        connection="dummy_out_conn",
                        data_type=DataType.STRING)
        def dummy():
            pass

        func = self._get_user_function(app)

        bindings = func.get_bindings()
        self.assertEqual(len(bindings), 2)

        output_binding = bindings[0]
        output = bindings[1]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.STRING,
            "type": BLOB_TRIGGER,
            "name": "req",
            "path": "dummy_path",
            "connection": "dummy_conn"
        })

        self.assertEqual(output_binding.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "dataType": DataType.STRING,
            "type": BLOB,
            "name": "out",
            "path": "dummy_out_path",
            "connection": "dummy_out_conn"
        })

    def test_custom_trigger(self):
        app = self.func_app

        @app.generic_trigger(arg_name="req", type=BLOB_TRIGGER,
                             data_type=DataType.BINARY,
                             connection="dummy_conn",
                             path="dummy_path")
        def dummy():
            pass

        func = self._get_user_function(app)

        bindings = func.get_bindings()
        self.assertEqual(len(bindings), 1)

        output = bindings[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.BINARY,
            "type": BLOB_TRIGGER,
            "name": "req",
            "path": "dummy_path",
            "connection": "dummy_conn"
        })

    def test_custom_input_binding(self):
        app = self.func_app

        @app.generic_trigger(arg_name="req", type=TIMER_TRIGGER,
                             data_type=DataType.BINARY,
                             schedule="dummy_schedule")
        @app.generic_input_binding(arg_name="file", type=BLOB,
                                   path="dummy_in_path",
                                   connection="dummy_in_conn",
                                   data_type=DataType.STRING)
        def dummy():
            pass

        func = self._get_user_function(app)

        bindings = func.get_bindings()
        self.assertEqual(len(bindings), 2)

        input_binding = bindings[0]
        trigger = bindings[1]

        self.assertEqual(trigger.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.BINARY,
            "type": TIMER_TRIGGER,
            "name": "req",
            "schedule": "dummy_schedule"
        })

        self.assertEqual(input_binding.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "dataType": DataType.STRING,
            "type": BLOB,
            "name": "file",
            "path": "dummy_in_path",
            "connection": "dummy_in_conn"
        })

    def test_custom_output_binding(self):
        app = self.func_app

        @app.generic_trigger(arg_name="req", type=QUEUE_TRIGGER,
                             queue_name="dummy_queue",
                             connection="dummy_conn")
        @app.generic_output_binding(arg_name="out", type=BLOB,
                                    path="dummy_out_path",
                                    connection="dummy_out_conn",
                                    data_type=DataType.STRING)
        def dummy():
            pass

        func = self._get_user_function(app)

        output_binding = func.get_bindings()[0]
        output = func.get_bindings()[1]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": QUEUE_TRIGGER,
            "name": "req",
            "queueName": "dummy_queue",
            "connection": "dummy_conn"
        })

        self.assertEqual(output_binding.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "dataType": DataType.STRING,
            "type": BLOB,
            "name": "out",
            "path": "dummy_out_path",
            "connection": "dummy_out_conn"
        })

    def test_custom_http_trigger(self):
        app = self.func_app

        @app.generic_trigger(arg_name="req", type=HTTP_TRIGGER)
        def dummy():
            pass

        func = self._get_user_function(app)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": HTTP_TRIGGER,
            "name": "req",
            "route": "dummy",
            "authLevel": AuthLevel.FUNCTION
        })

    def test_custom_binding_with_excluded_params(self):
        app = self.func_app

        @app.generic_trigger(arg_name="req", type=QUEUE_TRIGGER,
                             direction=BindingDirection.INOUT)
        def dummy():
            pass

        func = self._get_user_function(app)

        output = func.get_bindings()[0]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": QUEUE_TRIGGER,
            "name": "req"
        })

    def test_mixed_custom_and_supported_binding(self):
        app = self.func_app

        @app.queue_trigger(arg_name="req", queue_name="dummy_queue",
                           connection="dummy_conn")
        @app.generic_output_binding(arg_name="out", type=BLOB,
                                    path="dummy_out_path",
                                    connection="dummy_out_conn",
                                    data_type=DataType.STRING)
        def dummy():
            pass

        func = self._get_user_function(app)

        output_binding = func.get_bindings()[0]
        output = func.get_bindings()[1]

        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": QUEUE_TRIGGER,
            "name": "req",
            "queueName": "dummy_queue",
            "connection": "dummy_conn"
        })

        self.assertEqual(output_binding.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "dataType": DataType.STRING,
            "type": BLOB,
            "name": "out",
            "path": "dummy_out_path",
            "connection": "dummy_out_conn"
        })

    def test_event_grid_default_args(self):
        app = self.func_app

        @app.event_grid_trigger(arg_name="req")
        @app.write_event_grid(
            arg_name="res",
            topic_endpoint_uri="dummy_topic_endpoint_uri",
            topic_key_setting="dummy_topic_key_setting")
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func,
                    {"scriptFile": "function_app.py",
                     "bindings": [
                         {
                             "direction": BindingDirection.OUT,
                             "type": EVENT_GRID,
                             "name": "res",
                             "topicKeySetting": "dummy_topic_key_setting",
                             "topicEndpointUri": "dummy_topic_endpoint_uri"
                         },
                         {
                             "direction": BindingDirection.IN,
                             "type": EVENT_GRID_TRIGGER,
                             "name": "req"
                         }
                     ]
                     })

    def test_event_grid_full_args(self):
        app = self.func_app

        @app.event_grid_trigger(arg_name="req",
                                data_type=DataType.UNDEFINED,
                                dummy_field="dummy")
        @app.write_event_grid(
            arg_name="res",
            topic_endpoint_uri="dummy_topic_endpoint_uri",
            topic_key_setting="dummy_topic_key_setting",
            data_type=DataType.UNDEFINED,
            dummy_field="dummy"
        )
        def dummy():
            pass

        func = self._get_user_function(app)

        assert_json(self, func,
                    {"scriptFile": "function_app.py",
                     "bindings": [
                         {
                             "direction": BindingDirection.OUT,
                             "type": EVENT_GRID,
                             "name": "res",
                             "topicKeySetting": "dummy_topic_key_setting",
                             "topicEndpointUri": "dummy_topic_endpoint_uri",
                             'dummyField': 'dummy',
                             "dataType": DataType.UNDEFINED
                         },
                         {
                             "direction": BindingDirection.IN,
                             "type": EVENT_GRID_TRIGGER,
                             "name": "req",
                             'dummyField': 'dummy',
                             "dataType": DataType.UNDEFINED
                         }
                     ]
                     })

    def test_event_grid_trigger(self):
        app = self.func_app

        @app.event_grid_trigger(arg_name="req")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 1)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.IN,
            "type": EVENT_GRID_TRIGGER,
            "name": "req"
        })

    def test_event_grid_output_binding(self):
        app = self.func_app

        @app.event_grid_trigger(arg_name="req")
        @app.write_event_grid(
            arg_name="res",
            topic_endpoint_uri="dummy_topic_endpoint_uri",
            topic_key_setting="dummy_topic_key_setting")
        def dummy():
            pass

        func = self._get_user_function(app)

        self.assertEqual(len(func.get_bindings()), 2)

        output = func.get_bindings()[0]
        self.assertEqual(output.get_dict_repr(), {
            "direction": BindingDirection.OUT,
            "type": EVENT_GRID,
            "name": "res",
            "topicEndpointUri": "dummy_topic_endpoint_uri",
            "topicKeySetting": "dummy_topic_key_setting"
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from datetime import datetime
import unittest
from typing import List

import azure.functions as func
import azure.functions.eventgrid as azf_event_grid


class MyTestCase(unittest.TestCase):
    def test_eventgrid_input_type(self):
        check_input_type = azf_event_grid.EventGridEventInConverter.\
            check_input_type_annotation
        self.assertTrue(check_input_type(func.EventGridEvent))
        self.assertFalse(check_input_type(List[func.EventGridEvent]))
        self.assertFalse(check_input_type(str))
        self.assertFalse(check_input_type(bytes))

    def test_eventgrid_output_type(self):
        check_output_type = azf_event_grid.EventGridEventOutConverter.\
            check_output_type_annotation
        self.assertTrue(check_output_type(func.EventGridOutputEvent))
        self.assertTrue(check_output_type(List[func.EventGridOutputEvent]))
        self.assertTrue(check_output_type(str))
        self.assertTrue(check_output_type(bytes))
        self.assertTrue(check_output_type(List[str]))

    def test_eventgrid_decode(self):
        eventGridEvent = azf_event_grid.EventGridEventInConverter.decode(
            data=self._generate_single_eventgrid_datum(), trigger_metadata=None
        )
        self.assertEqual(
            eventGridEvent.id,
            "00010001-0001-0001-0001-000100010001")
        self.assertEqual(eventGridEvent.subject, "eventhubs/test")
        self.assertEqual(eventGridEvent.event_type, "captureFileCreated")
        self.assertEqual(eventGridEvent.topic, "/TestTopic/namespaces/test")
        self.assertIsNotNone(eventGridEvent.get_json())

    def test_eventgrid_decode_with_null_data(self):
        eventGridEvent = azf_event_grid.EventGridEventInConverter.decode(
            data=self._generate_single_eventgrid_datum(
                with_data=False), trigger_metadata=None)
        self.assertEqual(
            eventGridEvent.id,
            "00010001-0001-0001-0001-000100010001")
        self.assertEqual(eventGridEvent.subject, "eventhubs/test")
        self.assertEqual(eventGridEvent.event_type, "captureFileCreated")
        self.assertEqual(eventGridEvent.topic, "/TestTopic/namespaces/test")
        self.assertIsNone(eventGridEvent.get_json())

    def test_eventgrid_encode_with_str_data(self):
        example_data = self._generate_single_eventgrid_str()
        eventGridDatum = azf_event_grid.EventGridEventOutConverter.encode(
            example_data, expected_type=type(example_data))
        self.assertEqual(eventGridDatum.type, "string")

    def test_eventgrid_encode_with_bytes_data(self):
        example_data = self._generate_single_eventgrid_str(True)
        eventGridDatum = azf_event_grid.EventGridEventOutConverter.encode(
            example_data, expected_type=type(example_data))
        self.assertEqual(eventGridDatum.type, "bytes")

    def test_eventgrid_encode_with_EventGridData(self):
        example_data = self._generate_single_eventgrid_event()
        event_grid_datum = azf_event_grid.EventGridEventOutConverter.encode(
            example_data, expected_type=type(example_data))

        self.assertEqual(event_grid_datum.type, "json")

    def test_eventgrid_encode_with_multiple_EventGridData(self):
        example_data = self._generate_multiple_eventgrid_event()
        event_grid_datum = azf_event_grid.EventGridEventOutConverter.encode(
            example_data, expected_type=type(example_data))

        self.assertEqual(event_grid_datum.type, "json")

    @staticmethod
    def _generate_single_eventgrid_datum(with_data=True, datum_type='json'):
        datum_with_data = """
{
  "topic": "/TestTopic/namespaces/test",
  "subject": "eventhubs/test",
  "eventType": "captureFileCreated",
  "eventTime": "2017-07-14T23:10:27.7689666Z",
  "id": "00010001-0001-0001-0001-000100010001",
  "data": {
    "fileUrl": "https://test.blob.core.windows.net/debugging/testblob.txt",
    "fileType": "AzureBlockBlob",
    "partitionId": "1",
    "sizeInBytes": 0,
    "eventCount": 0,
    "firstSequenceNumber": -1,
    "lastSequenceNumber": -1,
    "firstEnqueueTime": "0001-01-01T00:00:00",
    "lastEnqueueTime": "0001-01-01T00:00:00"
  },
  "dataVersion": "",
  "metadataVersion": "1"
}
"""
        datum_without_data = """
{
  "topic": "/TestTopic/namespaces/test",
  "subject": "eventhubs/test",
  "eventType": "captureFileCreated",
  "eventTime": "2017-07-14T23:10:27.7689666Z",
  "id": "00010001-0001-0001-0001-000100010001",
  "dataVersion": "",
  "metadataVersion": "1"
}"""

        datum = datum_with_data if with_data else datum_without_data

        if datum_type == 'bytes':
            datum = datum.encode('utf-8')

        return func.meta.Datum(datum, datum_type)

    @staticmethod
    def _generate_single_eventgrid_event(with_date=True):
        return azf_event_grid.azf_eventgrid.EventGridOutputEvent(
            id="id",
            subject='subject',
            event_type='eventType',
            event_time=datetime.utcnow(),
            data={"tag1": "value1", "tag2": "value2"} if with_date else {},
            data_version='dataVersion',
        )

    @staticmethod
    def _generate_multiple_eventgrid_event(with_date=True):
        return [azf_event_grid.azf_eventgrid.EventGridOutputEvent(
            id="id1",
            subject='subject1',
            event_type='eventType1',
            event_time=datetime.utcnow(),
            data={"tag1": "value1", "tag2": "value2"} if with_date else {},
            data_version='dataVersion',
        ), azf_event_grid.azf_eventgrid.EventGridOutputEvent(
            id="id2",
            subject='subject2',
            event_type='eventType2',
            event_time=datetime.utcnow(),
            data={"tag1": "value1", "tag2": "value2"} if with_date else {},
            data_version='dataVersion',
        )]

    @staticmethod
    def _generate_single_eventgrid_str(in_bytes=False):
        string_representation = '{"id": "id", ' \
                                '"subject": "subject", ' \
                                '"dataVersion": "dataVersion", ' \
                                '"eventType": "eventType", ' \
                                '"data": {"tag1": "value1", ' \
                                '"tag2": "value2"}, ' \
                                '"eventTime": "2020-04-22T18:19:19Z"}'
        return string_representation.encode('utf-8') \
            if in_bytes \
            else string_representation


if __name__ == '__main__':
    unittest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


class FunctionExtensionException(Exception):
    """Excpetion emitted from Azure Functions Python Worker extension
    """
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import abc
import typing


class AbstractKafkaEvent(abc.ABC):

    @abc.abstractmethod
    def get_body(self) -> bytes:
        pass

    @property
    @abc.abstractmethod
    def key(self) -> typing.Optional[str]:
        pass

    @property
    @abc.abstractmethod
    def offset(self) -> typing.Optional[int]:
        pass

    @property
    @abc.abstractmethod
    def partition(self) -> typing.Optional[int]:
        pass

    @property
    @abc.abstractmethod
    def topic(self) -> typing.Optional[str]:
        pass

    @property
    @abc.abstractmethod
    def timestamp(self) -> typing.Optional[str]:
        pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import collections.abc
import json
import typing

from azure.functions import _cosmosdb as cdb

from . import meta


class CosmosDBConverter(meta.InConverter, meta.OutConverter,
                        binding='cosmosDB'):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, cdb.DocumentList)

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (cdb.DocumentList, cdb.Document))

    @classmethod
    def decode(cls,
               data: meta.Datum,
               *,
               trigger_metadata) -> typing.Optional[cdb.DocumentList]:
        if data is None or data.type is None:
            return None

        data_type = data.type

        if data_type in ['string', 'json']:
            body = data.value

        elif data_type == 'bytes':
            body = data.value.decode('utf-8')

        else:
            raise NotImplementedError(
                f'unsupported queue payload type: {data_type}')

        documents = json.loads(body)
        if not isinstance(documents, list):
            documents = [documents]

        return cdb.DocumentList(
            (None if doc is None else cdb.Document.from_dict(doc))
            for doc in documents)

    @classmethod
    def encode(cls, obj: typing.Any, *,
               expected_type: typing.Optional[type]) -> meta.Datum:
        if isinstance(obj, cdb.Document):
            data = cdb.DocumentList([obj])

        elif isinstance(obj, cdb.DocumentList):
            data = obj

        elif isinstance(obj, collections.abc.Iterable):
            data = cdb.DocumentList()

            for doc in obj:
                if not isinstance(doc, cdb.Document):
                    raise NotImplementedError
                else:
                    data.append(doc)

        else:
            raise NotImplementedError

        return meta.Datum(
            type='json',
            value=json.dumps([dict(d) for d in data])
        )


class CosmosDBTriggerConverter(CosmosDBConverter,
                               binding='cosmosDBTrigger', trigger=True):
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import azure.functions as func
from azure.functions._abc import TraceContext, RetryContext
from azure.functions._http_asgi import (
    AsgiMiddleware
)


class MockAsgiApplication:
    response_code = 200
    response_body = b''
    response_headers = [
        [b"content-type", b"text/plain"],
    ]

    async def __call__(self, scope, receive, send):
        self.received_scope = scope
        # Verify against ASGI specification
        assert scope['type'] == 'http'
        assert isinstance(scope['type'], str)

        assert scope['asgi.spec_version'] in ['2.0', '2.1']
        assert isinstance(scope['asgi.spec_version'], str)

        assert scope['asgi.version'] in ['2.0', '2.1', '2.2']
        assert isinstance(scope['asgi.version'], str)

        assert scope['http_version'] in ['1.0', '1.1', '2']
        assert isinstance(scope['http_version'], str)

        assert scope['method'] in ['POST', 'GET', 'PUT', 'DELETE', 'PATCH']
        assert isinstance(scope['method'], str)

        assert scope['scheme'] in ['http', 'https']
        assert isinstance(scope['scheme'], str)

        assert isinstance(scope['path'], str)
        assert isinstance(scope['raw_path'], bytes)
        assert isinstance(scope['query_string'], bytes)
        assert isinstance(scope['root_path'], str)

        assert hasattr(scope['headers'], '__iter__')
        for k, v in scope['headers']:
            assert isinstance(k, bytes)
            assert isinstance(v, bytes)

        assert scope['client'] is None or hasattr(scope['client'], '__iter__')
        if scope['client']:
            assert len(scope['client']) == 2
            assert isinstance(scope['client'][0], str)
            assert isinstance(scope['client'][1], int)

        assert scope['server'] is None or hasattr(scope['server'], '__iter__')
        if scope['server']:
            assert len(scope['server']) == 2
            assert isinstance(scope['server'][0], str)
            assert isinstance(scope['server'][1], int)

        self.received_request = await receive()
        assert self.received_request['type'] == 'http.request'
        assert isinstance(self.received_request['body'], bytes)
        assert isinstance(self.received_request['more_body'], bool)

        self.next_request = await receive()
        assert self.next_request['type'] == 'http.disconnect'

        await send(
            {
                "type": "http.response.start",
                "status": self.response_code,
                "headers": self.response_headers,
            }
        )
        await send(
            {
                "type": "http.response.body",
                "body": self.response_body,
            }
        )


class TestHttpAsgiMiddleware(unittest.TestCase):
    def _generate_func_request(
            self,
            method="POST",
            url="https://function.azurewebsites.net/api/http?firstname=rt",
            headers={
                "Content-Type": "application/json",
                "x-ms-site-restricted-token": "xmsrt"
            },
            params={
                "firstname": "roger"
            },
            route_params={},
            body=b'{ "lastname": "tsang" }'
    ) -> func.HttpRequest:
        return func.HttpRequest(
            method=method,
            url=url,
            headers=headers,
            params=params,
            route_params=route_params,
            body=body
        )

    def _generate_func_context(
        self,
        invocation_id='123e4567-e89b-12d3-a456-426655440000',
        function_name='httptrigger',
        function_directory='/home/roger/wwwroot/httptrigger',
        trace_context=TraceContext,
        retry_context=RetryContext
    ) -> func.Context:
        class MockContext(func.Context):
            def __init__(self, ii, fn, fd, tc, rc):
                self._invocation_id = ii
                self._function_name = fn
                self._function_directory = fd
                self._trace_context = tc
                self._retry_context = rc

            @property
            def invocation_id(self):
                return self._invocation_id

            @property
            def function_name(self):
                return self._function_name

            @property
            def function_directory(self):
                return self._function_directory

            @property
            def trace_context(self):
                return self._trace_context

            @property
            def retry_context(self):
                return self._retry_context

        return MockContext(invocation_id, function_name, function_directory,
                           trace_context, retry_context)

    def test_middleware_calls_app(self):
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = func.HttpRequest(method='get', url='/test', body=b'')
        response = AsgiMiddleware(app).handle(req)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_middleware_calls_app_with_context(self):
        """Test if the middleware can be used by exposing the .handle method,
        specifically when the middleware is used as
        def main(req, context):
            return AsgiMiddleware(app).handle(req, context)
        """
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = self._generate_func_request()
        ctx = self._generate_func_context()
        response = AsgiMiddleware(app).handle(req, ctx)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_middleware_wrapper(self):
        """Test if the middleware can be used by exposing the .main property,
        specifically when the middleware is used as
        main = AsgiMiddleware(app).main
        """
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = self._generate_func_request()
        ctx = self._generate_func_context()

        main = AsgiMiddleware(app).main
        response = main(req, ctx)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional, Union

from azure.functions.decorators.constants import COSMOS_DB, COSMOS_DB_TRIGGER
from azure.functions.decorators.core import DataType, InputBinding, \
    OutputBinding, Trigger


class CosmosDBInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return COSMOS_DB

    def __init__(self,
                 name: str,
                 database_name: str,
                 collection_name: str,
                 connection_string_setting: str,
                 data_type: Optional[DataType] = None,
                 id: Optional[str] = None,
                 sql_query: Optional[str] = None,
                 partition_key: Optional[str] = None,
                 **kwargs):
        self.database_name = database_name
        self.collection_name = collection_name
        self.connection_string_setting = connection_string_setting
        self.partition_key = partition_key
        self.id = id
        self.sql_query = sql_query
        super().__init__(name=name, data_type=data_type)


class CosmosDBOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return COSMOS_DB

    def __init__(self,
                 name: str,
                 database_name: str,
                 collection_name: str,
                 connection_string_setting: str,
                 create_if_not_exists: Optional[bool] = None,
                 collection_throughput: Optional[int] = None,
                 use_multiple_write_locations: Optional[bool] = None,
                 preferred_locations: Optional[str] = None,
                 partition_key: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.database_name = database_name
        self.collection_name = collection_name
        self.connection_string_setting = connection_string_setting
        self.create_if_not_exists = create_if_not_exists
        self.partition_key = partition_key
        self.collection_throughput = collection_throughput
        self.use_multiple_write_locations = use_multiple_write_locations
        self.preferred_locations = preferred_locations
        super().__init__(name=name, data_type=data_type)


class CosmosDBTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return COSMOS_DB_TRIGGER

    def __init__(self,
                 name: str,
                 database_name: str,
                 collection_name: str,
                 connection_string_setting: str,
                 leases_collection_throughput: Optional[int] = None,
                 checkpoint_interval: Optional[int] = None,
                 checkpoint_document_count: Optional[int] = None,
                 feed_poll_delay: Optional[int] = None,
                 lease_renew_interval: Optional[int] = None,
                 lease_acquire_interval: Optional[int] = None,
                 lease_expiration_interval: Optional[int] = None,
                 max_items_per_invocation: Optional[int] = None,
                 start_from_beginning: Optional[bool] = None,
                 create_lease_collection_if_not_exists: Optional[bool] = None,
                 preferred_locations: Optional[str] = None,
                 data_type: Optional[Union[DataType]] = None,
                 lease_collection_name: Optional[str] = None,
                 lease_connection_string_setting: Optional[str] = None,
                 lease_database_name: Optional[str] = None,
                 lease_collection_prefix: Optional[str] = None,
                 **kwargs):
        self.lease_collection_name = lease_collection_name
        self.lease_connection_string_setting = lease_connection_string_setting
        self.lease_database_name = lease_database_name
        self.create_lease_collection_if_not_exists = \
            create_lease_collection_if_not_exists
        self.leases_collection_throughput = leases_collection_throughput
        self.lease_collection_prefix = lease_collection_prefix
        self.checkpoint_interval = checkpoint_interval
        self.checkpoint_document_count = checkpoint_document_count
        self.feed_poll_delay = feed_poll_delay
        self.lease_renew_interval = lease_renew_interval
        self.lease_acquire_interval = lease_acquire_interval
        self.lease_expiration_interval = lease_expiration_interval
        self.max_items_per_invocation = max_items_per_invocation
        self.start_from_beginning = start_from_beginning
        self.preferred_locations = preferred_locations
        self.connection_string_setting = connection_string_setting
        self.database_name = database_name
        self.collection_name = collection_name
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import collections
import json

from . import _abc


class SqlRow(_abc.SqlRow, collections.UserDict):
    """A SQL Row.

    SqlRow objects are ''UserDict'' subclasses and behave like dicts.
    """

    @classmethod
    def from_json(cls, json_data: str) -> 'SqlRow':
        """Create a SqlRow from a JSON string."""
        return cls.from_dict(json.loads(json_data))

    @classmethod
    def from_dict(cls, dct: dict) -> 'SqlRow':
        """Create a SqlRow from a dict object"""
        return cls({k: v for k, v in dct.items()})

    def to_json(self) -> str:
        """Return the JSON representation of the SqlRow"""
        return json.dumps(dict(self))

    def __getitem__(self, key):
        return collections.UserDict.__getitem__(self, key)

    def __setitem__(self, key, value):
        return collections.UserDict.__setitem__(self, key, value)

    def __repr__(self) -> str:
        return (
            f'<SqlRow at 0x{id(self):0x}>'
        )


class SqlRowList(_abc.SqlRowList, collections.UserList):
    "A ''UserList'' subclass containing a list of :class:'~SqlRow' objects"
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import typing
from logging import Logger
from .extension_meta import ExtensionMeta
from .extension_scope import ExtensionScope
from .._abc import Context


class AppExtensionBase(metaclass=ExtensionMeta):
    """An abstract class defines the global life-cycle hooks to be implemented
    by customer's extension, will be applied to all functions.

    An AppExtension should be treated as a static class. Must not contain
    __init__ method since it is not instantiable.

    Please place your initialization code in init() classmethod, consider
    accepting extension settings in configure() classmethod from customers.
    """

    _scope = ExtensionScope.APPLICATION

    @classmethod
    def init(cls):
        """The function will be executed when the extension is loaded.
        Happens when Azure Functions customers import the extension module.
        """
        pass

    @classmethod
    def configure(cls, *args, **kwargs):
        """This function is intended to be called by Azure Functions
        customers. This is a contract between extension developers and
        azure functions customers. If multiple .configure() are called,
        the extension system cannot guarentee the calling order.
        """
        pass

    # DO NOT decorate this with @abc.abstractstatismethod
    # since implementation by subclass is not mandatory
    @classmethod
    def post_function_load_app_level(cls,
                                     function_name: str,
                                     function_directory: str,
                                     *args, **kwargs) -> None:
        """This must be implemented as a @classmethod. It will be called right
        a customer's function is loaded. In this stage, the customer's logger
        is not fully initialized from the Python worker. Please use print()
        to emit message if necessary.

        Parameters
        ----------
        function_name: str
            The name of customer's function (e.g. HttpTrigger)
        function_directory: str
            The path to customer's function directory
            (e.g. /home/site/wwwroot/HttpTrigger)
        """
        pass

    # DO NOT decorate this with @abc.abstractstatismethod
    # since implementation by subclass is not mandatory
    @classmethod
    def pre_invocation_app_level(cls,
                                 logger: Logger,
                                 context: Context,
                                 func_args: typing.Dict[str, object] = {},
                                 *args,
                                 **kwargs) -> None:
        """This must be implemented as a @staticmethod. It will be called right
        before a customer's function is being executed.

        Parameters
        ----------
        logger: logging.Logger
            A logger provided by Python worker. Extension developer should
            use this logger to emit telemetry to Azure Functions customers.
        context: azure.functions.Context
            This will include the function_name, function_directory and an
            invocation_id of this specific invocation.
        func_args: typing.Dict[str, object]
            Arguments that are passed into the Azure Functions. The name of
            each parameter is defined in function.json. Extension developers
            may also want to do isinstance() check if you want to apply
            operations to specific trigger types or input binding types.
        """
        pass

    # DO NOT decorate this with @abc.abstractstatismethod
    # since implementation by subclass is not mandatory
    @classmethod
    def post_invocation_app_level(cls,
                                  logger: Logger,
                                  context: Context,
                                  func_args: typing.Dict[str, object] = {},
                                  func_ret: typing.Optional[object] = None,
                                  *args,
                                  **kwargs) -> None:
        """This must be implemented as a @staticmethod. It will be called right
        before a customer's function is being executed.

        Parameters
        ----------
        logger: logging.Logger
            A logger provided by Python worker. Extension developer should
            use this logger to emit telemetry to Azure Functions customers.
        context: azure.functions.Context
            This will include the function_name, function_directory and an
            invocation_id of this specific invocation.
        func_args: typing.Dict[str, object]
            Arguments that are passed into the Azure Functions. The name of
            each parameter is defined in function.json. Extension developers
            may also want to do isinstance() check if you want to apply
            operations to specific trigger types or input binding types.
        func_ret: typing.Optional[object]
            Return value from Azure Functions. This is usually the value
            defined in function.json $return section. Extension developers
            may also want to do isinstance() check if you want to apply
            operations to specific types or input binding types."
        """
        pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, List, Tuple, Optional, Any, Union
import logging
import asyncio
from wsgiref.headers import Headers

from ._abc import Context
from ._http import HttpRequest, HttpResponse
from ._http_wsgi import WsgiRequest


class AsgiRequest(WsgiRequest):
    def __init__(self, func_req: HttpRequest,
                 func_ctx: Optional[Context] = None):
        self.asgi_version = "2.1"
        self.asgi_spec_version = "2.1"
        self._headers = func_req.headers
        super().__init__(func_req, func_ctx)

    def _get_encoded_http_headers(self) -> List[Tuple[bytes, bytes]]:
        return [(k.encode("utf8"), v.encode("utf8"))
                for k, v in self._headers.items()]

    def _get_server_address(self):
        if self.server_name is not None:
            return (self.server_name, int(self.server_port))
        return None

    def to_asgi_http_scope(self):
        if self.path_info is not None:
            _raw_path = self.path_info.encode("utf-8")
        else:
            _raw_path = b''
        if self.query_string is not None:
            _query_string = self.query_string.encode("utf-8")
        else:
            _query_string = b''

        return {
            "type": "http",
            "asgi.version": self.asgi_version,
            "asgi.spec_version": self.asgi_spec_version,
            "http_version": "1.1",
            "method": self.request_method,
            "scheme": "https",
            "path": self.path_info,
            "raw_path": _raw_path,
            "query_string": _query_string,
            "root_path": self.script_name,
            "headers": self._get_encoded_http_headers(),
            "server": self._get_server_address(),
            "client": None,
            "azure_functions.function_directory": self.af_function_directory,
            "azure_functions.function_name": self.af_function_name,
            "azure_functions.invocation_id": self.af_invocation_id,
            "azure_functions.trace_context": self.af_trace_context,
            "azure_functions.retry_context": self.af_retry_context
        }
        # Notes, missing client name, port


class AsgiResponse:
    def __init__(self):
        self._status_code = 0
        self._headers: Union[Headers, Dict] = {}
        self._buffer: List[bytes] = []
        self._request_body: Optional[bytes] = b""

    @classmethod
    async def from_app(cls, app, scope: Dict[str, Any],
                       body: bytes) -> "AsgiResponse":
        res = cls()
        res._request_body = body
        await app(scope, res._receive, res._send)
        return res

    def to_func_response(self) -> HttpResponse:
        lowercased_headers = {k.lower(): v for k, v in self._headers.items()}
        return HttpResponse(
            body=b"".join(self._buffer),
            status_code=self._status_code,
            headers=self._headers,  # type: ignore
            mimetype=lowercased_headers.get("content-type"),
            charset=lowercased_headers.get("content-encoding"),
        )

    def _handle_http_response_start(self, message: Dict[str, Any]):
        self._headers = Headers(
            [(k.decode(), v.decode())
             for k, v in message["headers"]])
        self._status_code = message["status"]

    def _handle_http_response_body(self, message: Dict[str, Any]):
        self._buffer.append(message["body"])
        # XXX : Chunked bodies not supported, see
        # https://github.com/Azure/azure-functions-host/issues/4926

    async def _receive(self):
        if self._request_body is not None:
            reply = {
                "type": "http.request",
                "body": self._request_body,
                "more_body": False,
            }
            self._request_body = None
        else:
            reply = {
                "type": "http.disconnect",
            }
        return reply

    async def _send(self, message):
        logging.debug(f"Received {message} from ASGI worker.")
        if message["type"] == "http.response.start":
            self._handle_http_response_start(message)
        elif message["type"] == "http.response.body":
            self._handle_http_response_body(message)
        elif message["type"] == "http.disconnect":
            pass  # Nothing todo here


class AsgiMiddleware:
    """This middleware is to adapt an ASGI supported Python server
    framework into Azure Functions. It can be used by either calling the
    .handle() function or exposing the .main property in a HttpTrigger.
    """
    _logger = logging.getLogger('azure.functions.AsgiMiddleware')
    _usage_reported = False

    def __init__(self, app):
        """Instantiate an ASGI middleware to convert Azure Functions HTTP
        request into ASGI Python object. Example on handling ASGI app in a HTTP
        trigger by overwriting the .main() method:

        import azure.functions as func

        from FastapiApp import app

        main = func.AsgiMiddleware(app).main
        """
        if not self._usage_reported:
            self._logger.info("Instantiating Azure Functions ASGI middleware.")
            self._usage_reported = True

        self._app = app
        self._loop = asyncio.new_event_loop()
        self.main = self._handle

    def handle(self, req: HttpRequest, context: Optional[Context] = None):
        """Method to convert an Azure Functions HTTP request into a ASGI
        Python object. Example on handling ASGI app in a HTTP trigger by
        calling .handle() in .main() method:

        import azure.functions as func

        from FastapiApp import app

        def main(req, context):
            return func.AsgiMiddleware(app).handle(req, context)
        """
        self._logger.debug(f"Handling {req.url} as an ASGI request.")
        return self._handle(req, context)

    def _handle(self, req, context):
        asgi_request = AsgiRequest(req, context)
        asyncio.set_event_loop(self._loop)
        scope = asgi_request.to_asgi_http_scope()
        asgi_response = self._loop.run_until_complete(
            AsgiResponse.from_app(self._app, scope, req.get_body())
        )

        return asgi_response.to_func_response()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from setuptools import find_packages, setup
from azure.functions import __version__

EXTRA_REQUIRES = {
    'dev': [
        'flake8~=4.0.1',
        'mypy',
        'pytest',
        'pytest-cov',
        'requests==2.*',
        'coverage'
    ]
}

with open("README.md") as readme:
    long_description = readme.read()

setup(
    name='azure-functions',
    version=__version__,
    description='Azure Functions for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Microsoft Corporation',
    author_email='azpysdkhelp@microsoft.com',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Environment :: Web Environment',
        'Development Status :: 5 - Production/Stable',
    ],
    license='MIT',
    packages=find_packages(exclude=[
        'azure'
    ]),
    package_data={
        'azure.functions': ['py.typed']
    },
    extras_require=EXTRA_REQUIRES,
    include_package_data=True,
    test_suite='tests'
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import sys
import unittest
from unittest import skipIf

import azure.functions as func
import azure.functions.http as http
from azure.functions._http import HttpResponseHeaders


class TestHTTP(unittest.TestCase):
    def test_http_form_parse_files(self):
        data = (
            b"--foo\r\n"
            b"Content-Type: text/plain; charset=utf-8\r\n"
            b"Content-Disposition: form-data; name=rfc2231;\r\n"
            b"  filename*0*=ascii''a%20b%20;\r\n"
            b"  filename*1*=c%20d%20;\r\n"
            b'  filename*2="e f.txt"\r\n\r\n'
            b"file contents\r\n--foo--"
        )
        request = func.HttpRequest(
            method='POST',
            url='/foo',
            body=data,
            headers={
                'Content-Type': 'multipart/form-data; boundary=foo'
            }
        )

        self.assertEqual(request.files["rfc2231"].filename, "a b c d e f.txt")
        self.assertEqual(request.files["rfc2231"].read(), b"file contents")

    def test_http_json(self):
        data: bytes = b'{ "result": "OK" }'
        request = func.HttpRequest(
            method='POST',
            url='/foo',
            body=data,
            headers={
                'Content-Type': 'application/json; charset=utf-8'
            }
        )

        self.assertEqual(request.get_body(), b'{ "result": "OK" }')
        self.assertEqual(request.get_json().get('result'), 'OK')

    def test_http_form_parse_urlencoded(self):
        data = b"foo=Hello+World&bar=baz"
        req = func.HttpRequest(
            method="POST",
            url='/foo',
            body=data,
            headers={
                'Content-Type': 'application/x-www-form-urlencoded'
            },
        )
        self.assertEqual(req.form["foo"], u"Hello World")

    def test_http_form_parse_formdata(self):
        data = (
            b"--foo\r\nContent-Disposition: form-field; name=foo\r\n\r\n"
            b"Hello World\r\n"
            b"--foo\r\nContent-Disposition: form-field; name=bar\r\n\r\n"
            b"bar=baz\r\n--foo--"
        )
        req = func.HttpRequest(
            method="POST",
            url='/foo',
            body=data,
            headers={
                'Content-Type': 'multipart/form-data; boundary=foo'
            },
        )
        self.assertEqual(req.form["foo"], u"Hello World")
        self.assertEqual(req.form["bar"], u"bar=baz")

    def test_http_input_type(self):
        check_input_type = (
            http.HttpRequestConverter.check_input_type_annotation
        )
        self.assertTrue(check_input_type(func.HttpRequest))
        self.assertFalse(check_input_type(str))

    def test_http_output_type(self):
        check_output_type = (
            http.HttpResponseConverter.check_output_type_annotation
        )
        self.assertTrue(check_output_type(func.HttpResponse))
        self.assertTrue(check_output_type(str))

    def test_http_response_encode_to_datum_no_cookie(self):
        resp = func.HttpResponse()
        datum = http.HttpResponseConverter.encode(resp, expected_type=None)

        self.assertEqual(datum.value["cookies"], None)

    @skipIf(sys.version_info < (3, 8, 0),
            "Skip the tests for Python 3.7 and below")
    def test_http_response_encode_to_datum_with_cookies(self):
        headers = HttpResponseHeaders()
        headers.add("Set-Cookie",
                    'foo3=42; Domain=example.com; Expires=Thu, '
                    '12-Jan-2017 13:55:08 GMT; Path=/; Max-Age=10000000')
        headers.add("Set-Cookie",
                    'foo3=43; Domain=example.com; Expires=Thu, 12-Jan-2018 '
                    '13:55:09 GMT; Path=/; Max-Age=10000000')
        resp = func.HttpResponse(headers=headers)
        datum = http.HttpResponseConverter.encode(resp, expected_type=None)

        actual_cookies = datum.value['cookies']
        self.assertIsNotNone(actual_cookies)
        self.assertTrue(isinstance(actual_cookies, list))
        self.assertTrue(len(actual_cookies), 2)
        self.assertEqual(str(actual_cookies[0]),
                         "Set-Cookie: foo3=42; Domain=example.com; "
                         "expires=Thu, 12-Jan-2017 13:55:08 GMT; "
                         "Max-Age=10000000; Path=/")
        self.assertEqual(str(actual_cookies[1]),
                         "Set-Cookie: foo3=43; Domain=example.com; "
                         "expires=Thu, 12-Jan-2018 13:55:09 GMT; "
                         "Max-Age=10000000; Path=/")

        self.assertTrue("Set-Cookie" not in resp.headers)

    @skipIf(sys.version_info >= (3, 8, 0),
            "Skip the tests for Python 3.8 and above")
    def test_http_response_encode_to_datum_with_cookies_in_python_3_7(self):
        headers = HttpResponseHeaders()
        headers.add("Set-Cookie",
                    'foo3=42; Domain=example.com; Expires=Thu, '
                    '12-Jan-2017 13:55:08 GMT; Path=/; Max-Age=10000000')
        headers.add("Set-Cookie",
                    'foo3=43; Domain=example.com; Expires=Thu, 12-Jan-2018 '
                    '13:55:09 GMT; Path=/; Max-Age=10000000')
        resp = func.HttpResponse(headers=headers)
        datum = http.HttpResponseConverter.encode(resp, expected_type=None)

        actual_cookies = datum.value['cookies']
        self.assertIsNone(actual_cookies)
        self.assertIn("Set-Cookie", resp.headers,
                      "Set-Cookie header not present in response headers!")

    @skipIf(sys.version_info < (3, 8, 0),
            "Skip the tests for Python 3.7 and below")
    def test_http_response_encode_to_datum_with_cookies_lower_case(self):
        headers = HttpResponseHeaders()
        headers.add("set-cookie",
                    'foo3=42; Domain=example.com; Path=/; Max-Age=10000.0')
        resp = func.HttpResponse(headers=headers)
        datum = http.HttpResponseConverter.encode(resp, expected_type=None)

        actual_cookies = datum.value['cookies']
        self.assertIsNotNone(actual_cookies)
        self.assertTrue(isinstance(actual_cookies, list))
        self.assertTrue(len(actual_cookies), 1)
        self.assertEqual(str(actual_cookies[0]),
                         "Set-Cookie: foo3=42; Domain=example.com; "
                         "Max-Age=10000.0; Path=/")

    def test_http_request_should_not_have_implicit_output(self):
        self.assertFalse(http.HttpRequestConverter.has_implicit_output())

    def test_http_response_does_not_have_explicit_output(self):
        self.assertIsNone(
            getattr(http.HttpResponseConverter, 'has_implicit_output', None)
        )



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import BLOB_TRIGGER, BLOB
from azure.functions.decorators.core import Trigger, OutputBinding, DataType, \
    InputBinding


class BlobTrigger(Trigger):
    def __init__(self,
                 name: str,
                 path: str,
                 connection: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.path = path
        self.connection = connection
        super().__init__(name=name, data_type=data_type)

    @staticmethod
    def get_binding_name() -> str:
        return BLOB_TRIGGER


class BlobInput(InputBinding):
    def __init__(self,
                 name: str,
                 path: str,
                 connection: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.path = path
        self.connection = connection
        super().__init__(name=name, data_type=data_type)

    @staticmethod
    def get_binding_name() -> str:
        return BLOB


class BlobOutput(OutputBinding):
    def __init__(self,
                 name: str,
                 path: str,
                 connection: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.path = path
        self.connection = connection
        super().__init__(name=name, data_type=data_type)

    @staticmethod
    def get_binding_name() -> str:
        return BLOB
